# -*- coding: utf-8 -*-
from cone.app import cfg
from cone.app import testing
from cone.app.compat import configparser
from cone.app.compat import StringIO
from cone.app.interfaces import ILeafNode
from cone.app.interfaces import IMetadata
from cone.app.interfaces import INodeInfo
from cone.app.interfaces import IProperties
from cone.app.model import AdapterNode
from cone.app.model import AppEnvironment
from cone.app.model import AppNode
from cone.app.model import BaseNode
from cone.app.model import ConfigProperties
from cone.app.model import FactoryNode
from cone.app.model import get_node_info
from cone.app.model import LanguageSchema
from cone.app.model import LeafNode
from cone.app.model import Metadata
from cone.app.model import NamespaceUUID
from cone.app.model import node_info
from cone.app.model import NodeInfo
from cone.app.model import o_getattr
from cone.app.model import Properties
from cone.app.model import ProtectedProperties
from cone.app.model import register_node_info
from cone.app.model import Translation
from cone.app.model import UUIDAsName
from cone.app.model import UUIDAttributeAware
from cone.app.model import XMLProperties
from datetime import datetime
from node import schema
from node.behaviors import DefaultInit
from node.behaviors import DictStorage
from node.behaviors import MappingAdopt
from node.behaviors import MappingConstraints
from node.behaviors import MappingNode
from node.behaviors import Node
from node.behaviors import NodeInit
from node.interfaces import IInvalidate
from node.interfaces import ILeaf
from node.tests import NodeTestCase
from odict import odict
from plumber import plumbing
from pyramid.security import ALL_PERMISSIONS
from zope.component.globalregistry import BaseGlobalComponents
import copy
import os
import shutil
import tempfile
import uuid


class TestModel(NodeTestCase):
    layer = testing.security

    def test_BaseNode(self):
        # Default permissions.
        root = BaseNode()
        self.assertEqual(root.__acl__, [
            ('Allow', 'system.Authenticated', ['view']),
            ('Allow', 'role:viewer', ['view', 'list']),
            ('Allow', 'role:editor', [
                'view', 'list', 'add', 'edit', 'change_order'
            ]),
            ('Allow', 'role:admin', [
                'view', 'list', 'add', 'edit', 'change_order', 'delete', 'cut',
                'copy', 'paste', 'manage_permissions', 'change_state'
            ]),
            ('Allow', 'role:manager', [
                'view', 'list', 'add', 'edit', 'change_order', 'delete', 'cut',
                'copy', 'paste', 'manage_permissions', 'change_state', 'manage'
            ]),
            ('Allow', 'role:owner', [
                'view', 'list', 'add', 'edit', 'change_order', 'delete', 'cut',
                'copy', 'paste', 'manage_permissions', 'change_state'
            ]),
            ('Allow', 'system.Everyone', ['login']),
            ('Deny', 'system.Everyone', ALL_PERMISSIONS)
        ])

        # Properties
        props = root.properties
        self.assertTrue(isinstance(props, Properties))
        self.assertTrue(IProperties.providedBy(props))

        # ``IProperties`` implementations do not raise AttributeErrors if
        # requested attribute not exists
        self.assertTrue(props.inexistent is None)

        # Metadata
        md = root.metadata
        self.assertTrue(isinstance(md, Metadata))
        self.assertTrue(IMetadata.providedBy(md))
        self.assertTrue(IProperties.providedBy(md))

        self.assertEqual(md.title, 'no_title')
        self.assertTrue(md.inexistent is None)

        # Nodeinfo
        info = root.nodeinfo
        self.assertTrue(isinstance(info, NodeInfo))
        self.assertTrue(INodeInfo.providedBy(info))
        self.assertTrue(info.node is BaseNode)
        self.assertEqual(info.title, "<class 'cone.app.model.BaseNode'>")
        self.assertTrue(info.inexistent is None)

    def test_LeafNode(self):
        @plumbing(NodeInit, Node, LeafNode)
        class TestLeafNode(object):
            pass

        node = TestLeafNode()
        self.assertTrue(ILeaf.providedBy(node))
        self.assertTrue(ILeafNode.providedBy(node))
        with self.assertRaises(KeyError):
            node['child']
        with self.assertRaises(KeyError):
            node['child'] = BaseNode()
        with self.assertRaises(KeyError):
            del node['child']
        self.assertEqual(list(node), [])

    def test_FactoryNode(self):
        class TestFactoryNode(FactoryNode):
            factories = odict()
            factories['foo'] = BaseNode
            factories['bar'] = BaseNode
        node = TestFactoryNode()

        # static factories
        self.assertTrue(str(node['foo']).startswith("<BaseNode object 'foo' at"))
        self.assertTrue(str(node['bar']).startswith("<BaseNode object 'bar' at"))
        self.expect_error(KeyError, lambda: node['baz'])
        self.assertEqual(list([it for it in node]), ['foo', 'bar'])
        self.assertTrue(IInvalidate.providedBy(node))
        self.assertEqual(node.values(), [node['foo'], node['bar']])

        # invalidate
        node.invalidate()
        self.assertEqual(node.storage.values(), [])
        self.assertEqual(node.values(), [node['foo'], node['bar']])
        self.assertEqual(node.storage.values(), [node['foo'], node['bar']])

        node.invalidate('foo')
        self.assertEqual(node.storage.values(), [node['bar']])

    def test_AdapterNode(self):
        toadapt = BaseNode()
        toadapt['foo'] = BaseNode()
        toadapt['bar'] = BaseNode()
        toadapt.attrs.title = 'Some title'
        toadapt.attrs.description = 'Some description'

        # Adapt created node structure
        adapter = AdapterNode(toadapt, 'name', None)

        # Check ``AdapterNode``
        expected = "<BaseNode object 'foo' at"
        self.assertTrue(str(adapter.model['foo']).startswith(expected))

        # ``attrs``
        self.assertEqual(adapter.attrs.title, 'Some title')

        # The adapter node is responsible to return other adapter node or
        # application nodes on ``__getitem__`` if application hierarchy
        # continues. It's possible to do key aliasing as well at this place.
        # This dummy class does a static mapping on ``__getitem__``.
        class TestAdapterNode(AdapterNode):
            def __getitem__(self, key):
                return AdapterNode(self.model['bar'], key, self)

        node = TestAdapterNode(toadapt, 'adapter', None)
        child = node['aliased']
        expected = "<AdapterNode object 'aliased' at"
        self.assertTrue(str(child).startswith(expected))

        self.assertTrue(child.model is toadapt['bar'])
        self.assertEqual([key for key in node], ['foo', 'bar'])

        # The application node path differs from the adapted node path. This is
        # essential to keep the application path sane while not violating the
        # adapted node's structure
        self.assertEqual(child.path, ['adapter', 'aliased'])
        self.assertEqual(child.model.path, [None, 'bar'])

    def test_Metadata(self):
        # The ``IMetadata`` implementation returned by
        # ``IApplicationNode.metadata`` is used by the application for
        # displaying metadata information. The default implementation accepts a
        # dict like object on ``__init__``
        data = {
            'title': 'some title',
            'description': 'some description',
            'creator': 'john doe',
        }

        # Check ``INodeAdapter`` interface
        metadata = Metadata(data)

        # ``__getattr__``. No AttributeError is raised if attribute is
        # inexistent
        self.assertEqual(metadata.title, 'some title')
        self.assertEqual(metadata.description, 'some description')
        self.assertEqual(metadata.creator, 'john doe')
        self.assertTrue(metadata.inexistent is None)

        # ``__getitem__``
        self.assertEqual(metadata['title'], 'some title')

        # ``__contains__``
        self.assertTrue('description' in metadata)

        # ``get``
        self.assertEqual(metadata.get('creator'), 'john doe')

        # internal data
        self.assertEqual(o_getattr(metadata, '_data'), {
            'creator': 'john doe',
            'description': 'some description',
            'title': 'some title'
        })

        # ``__copy__``
        metadata_copy = copy.copy(metadata)
        self.assertFalse(metadata_copy is metadata)
        self.assertEqual(metadata_copy.__class__, Metadata)
        self.assertEqual(
            o_getattr(metadata_copy, '_data'),
            o_getattr(metadata, '_data')
        )

        # ``__deepcopy__``
        metadata_deepcopy = copy.deepcopy(metadata)
        self.assertFalse(metadata_deepcopy is metadata)
        self.assertEqual(metadata_deepcopy.__class__, Metadata)
        self.assertEqual(
            o_getattr(metadata_deepcopy, '_data'),
            o_getattr(metadata, '_data')
        )

    @testing.reset_node_info_registry
    def test_NodeInfo(self):
        # The ``INodeInfo`` providing object holds information about the
        # application node
        nodeinfo = NodeInfo()
        nodeinfo.node = BaseNode
        nodeinfo.addables = ['basenode']
        nodeinfo.title = 'Base Node'
        nodeinfo.description = 'Base Node Description'
        nodeinfo.factory = None
        nodeinfo.icon = 'base-node-icon'

        # check interface
        self.assertTrue(INodeInfo.providedBy(nodeinfo))

        # check superclass
        self.assertTrue(isinstance(nodeinfo, Properties))

        # Register node info
        register_node_info('basenode', nodeinfo)

        # Lookup Node info
        nodeinfo = get_node_info('basenode')
        self.assertTrue(nodeinfo.node is BaseNode)
        self.assertEqual(nodeinfo.addables, ['basenode'])
        self.assertEqual(nodeinfo.title, 'Base Node')
        self.assertEqual(nodeinfo.description, 'Base Node Description')
        self.assertEqual(nodeinfo.factory, None)
        self.assertEqual(nodeinfo.icon, 'base-node-icon')

        # ``__getattr__``. No AttributeError is raised if attribute missing
        self.assertTrue(nodeinfo.inexistent is None)

        # ``__getitem__``
        self.assertEqual(nodeinfo['addables'], ['basenode'])

        # ``__contains__``
        self.assertTrue('node' in nodeinfo)

        # ``get``
        self.assertTrue(nodeinfo.get('node') is BaseNode)

        # internal data
        self.assertEqual(o_getattr(nodeinfo, '_data'), {
            'addables': ['basenode'],
            'description': 'Base Node Description',
            'factory': None,
            'icon': 'base-node-icon',
            'node': BaseNode,
            'title': 'Base Node'
        })

        # ``__copy__``
        nodeinfo_copy = copy.copy(nodeinfo)
        self.assertFalse(nodeinfo_copy is nodeinfo)
        self.assertEqual(nodeinfo_copy.__class__, NodeInfo)
        self.assertEqual(
            o_getattr(nodeinfo_copy, '_data'),
            o_getattr(nodeinfo, '_data')
        )

        # ``__deepcopy__``
        nodeinfo_deepcopy = copy.deepcopy(nodeinfo)
        self.assertFalse(nodeinfo_deepcopy is nodeinfo)
        self.assertEqual(nodeinfo_deepcopy.__class__, NodeInfo)
        self.assertEqual(
            o_getattr(nodeinfo_deepcopy, '_data'),
            o_getattr(nodeinfo, '_data')
        )

    @testing.reset_node_info_registry
    def test_node_info(self):
        @node_info(
            name='mynode',
            title='My Node',
            description='My Node Descriptrion',
            factory=None,
            icon='icon',
            addables=['othernode'])
        class MyNode(BaseNode):
            pass

        info = get_node_info('mynode')
        self.assertTrue(info.node is MyNode)
        self.assertEqual(info.title, 'My Node')
        self.assertEqual(info.description, 'My Node Descriptrion')
        self.assertEqual(info.factory, None)
        self.assertEqual(info.addables, ['othernode'])
        self.assertEqual(info.icon, 'icon')
        self.assertEqual(MyNode.node_info_name, 'mynode')

    def test_AppEnvironment(self):
        @plumbing(AppEnvironment)
        class AppEnvironmentNode(BaseNode):
            pass

        node = AppEnvironmentNode()

        # Request
        self.layer.forget_request()
        self.assertTrue(node.request is None)
        request = self.layer.new_request()
        self.assertTrue(node.request is request)

        # Registry
        self.assertIsInstance(node.registry, BaseGlobalComponents)

    def test_NamespaceUUID(self):
        @plumbing(NamespaceUUID)
        class NamespaceUUIDNode(BaseNode):
            pass

        node = NamespaceUUIDNode()
        self.assertEqual(node.uuid, None)

        with self.assertRaises(NotImplementedError):
            node.uuid = uuid.uuid4()

        root = BaseNode()
        root['ns_node'] = node
        self.assertEqual(
            node.uuid,
            uuid.UUID('44b1c59f-9de5-55d9-8e6f-4d6390be6ecd')
        )

        node.uuid_namespace = uuid.UUID('d737ada9-d400-486a-8795-57fedf05bb9f')
        self.assertEqual(
            node.uuid,
            uuid.UUID('c3861feb-bb1d-542d-a18f-e864146fea4d')
        )

    def test_UUIDAttributeAware(self):
        @plumbing(UUIDAttributeAware)
        class UUIDNode(BaseNode):
            pass

        node = UUIDNode()
        self.assertTrue(isinstance(node.uuid, uuid.UUID))
        self.assertTrue(node.attrs['uuid'] is node.uuid)

        UUIDNode.uuid_attribute_name = 'id'
        node = UUIDNode()
        self.assertTrue(isinstance(node.uuid, uuid.UUID))
        self.assertTrue(node.attrs['id'] is node.uuid)

    def test_UUIDAsName(self):
        @plumbing(UUIDAsName)
        class UUIDAsNameNode(BaseNode):
            pass

        node = UUIDAsNameNode()
        self.assertTrue(isinstance(node.uuid, uuid.UUID))
        self.assertTrue(str(node.uuid) == node.name)

        child = UUIDAsNameNode()
        node[child.name] = child
        sub = UUIDAsNameNode()
        node[child.name][sub.name] = sub
        sub = UUIDAsNameNode()
        node[child.name][sub.name] = sub
        self.check_output("""\
        <class '...UUIDAsNameNode'>: ...
          <class '...UUIDAsNameNode'>: ...
            <class '...UUIDAsNameNode'>: ...
            <class '...UUIDAsNameNode'>: ...
        """, node.treerepr())

        err = self.expect_error(RuntimeError, node[child.name].copy)
        expected = 'Shallow copy useless on UUID aware node trees, use deepcopy.'
        self.assertEqual(str(err), expected)

        copy = child.deepcopy()
        self.check_output("""\
        <class '...UUIDAsNameNode'>: ...
          <class '...UUIDAsNameNode'>: ...
          <class '...UUIDAsNameNode'>: ...
        """, copy.treerepr())

        self.assertFalse(copy.uuid == child.uuid)
        self.assertFalse(sorted(copy.keys()) == sorted(child.keys()))
        self.assertEqual(len(copy.keys()), 2)
        self.assertEqual(len(copy.values()), 2)
        self.assertTrue(copy[copy.keys()[0]].name == copy.keys()[0])

        @plumbing(
            AppNode,
            MappingConstraints,
            MappingAdopt,
            DefaultInit,
            MappingNode,
            DictStorage,
            UUIDAsName)
        class UnorderedUUIDAsNameNode(object):
            pass

        node = UUIDAsNameNode()
        child = UnorderedUUIDAsNameNode()
        node[child.name] = child
        copy = node.deepcopy()

        self.assertFalse(copy.uuid == node.uuid)
        self.assertFalse(copy[copy.keys()[0]] == child.uuid)

    def test_Translation(self):
        self.assertEqual(cfg.available_languages, ['en', 'de'])
        ls = LanguageSchema()
        self.assertEqual(ls.keys(), ['en', 'de'])
        self.assertIsInstance(ls['en'], schema.Str)
        self.assertTrue(ls.get('de') is not None)
        self.assertTrue(ls.get('it') is None)
        with self.assertRaises(KeyError):
            ls['it']

        @plumbing(Translation)
        class TranslationNode(BaseNode):
            child_constraints = None

        translation = TranslationNode(name='translation')
        self.assertEqual(translation.value, 'translation')
        self.assertEqual(translation.canonical_value, 'translation')

        translation['en'] = u'English Translation'
        translation['de'] = u'German Translation'
        self.assertEqual(translation['en'], 'English Translation')
        self.assertEqual(translation['de'], 'German Translation')

        self.layer.forget_request()
        self.assertEqual(translation.value, 'English Translation')
        self.assertEqual(translation.canonical_value, 'English Translation')

        self.layer.set_lang('de')
        self.assertEqual(translation.value, 'German Translation')
        self.assertEqual(translation.canonical_value, 'English Translation')

        self.layer.set_lang('en')
        self.assertEqual(translation.value, 'English Translation')
        self.assertEqual(translation.canonical_value, 'English Translation')

        self.layer.set_lang('it')
        self.assertEqual(translation.value, 'translation')
        self.assertEqual(translation.canonical_value, 'English Translation')

    def test_Properties(self):
        # ``Properties`` object can be used for any kind of mapping.
        props = Properties()

        props['foo'] = 'foo'
        self.assertEqual(props['foo'], 'foo')
        self.assertEqual(props.get('bar', default='default'), 'default')

        props.bar = 'bar'
        self.assertEqual(props.bar, 'bar')
        self.assertTrue('bar' in props)
        self.assertEqual(sorted(props.keys()), ['bar', 'foo'])

        self.assertEqual(o_getattr(props, '_data'), {'foo': 'foo', 'bar': 'bar'})

        props_copy = copy.copy(props)
        self.assertFalse(props_copy is props)
        self.assertEqual(props_copy.__class__, Properties)
        self.assertEqual(
            o_getattr(props_copy, '_data'),
            o_getattr(props, '_data')
        )

        props_deepcopy = copy.deepcopy(props)
        self.assertFalse(props_deepcopy is props)
        self.assertEqual(props_deepcopy.__class__, Properties)
        self.assertEqual(
            o_getattr(props_deepcopy, '_data'),
            o_getattr(props, '_data')
        )

    def test_ProtectedProperties(self):
        # Protected properties checks against permission for properties
        context = BaseNode()

        # 'viewprotected' property gets protected by 'view' permission
        permissions = {
            'viewprotected': ['view'],
        }
        props = ProtectedProperties(context, permissions)
        self.assertTrue(IProperties.providedBy(props))

        # Setting properties works always
        props.viewprotected = True
        props.unprotected = True

        # Unauthorized just permits access to unprotected property
        self.layer.new_request()

        self.assertTrue(props.viewprotected is None)
        self.assertTrue(props.unprotected)

        self.assertFalse('viewprotected' in props)
        self.assertTrue('unprotected' in props)

        self.assertEqual(props.keys(), ['unprotected'])
        self.assertTrue(props.get('viewprotected') is None)
        self.assertTrue(props.get('unprotected'))

        err = self.expect_error(KeyError, lambda: props['viewprotected'])
        expected = '"No permission to access \'viewprotected\'"'
        self.assertEqual(str(err).strip('u'), expected)

        self.assertTrue(props['unprotected'])

        # Authenticate, both properties are now available
        with self.layer.authenticated('viewer'):
            self.layer.new_request()

            self.assertTrue(props['viewprotected'])
            self.assertTrue(props.viewprotected)
            self.assertTrue(props.unprotected)

            self.assertEqual(
                sorted(props.keys()),
                ['unprotected', 'viewprotected']
            )
            self.assertTrue(props.get('viewprotected'))
            self.assertTrue(props.get('unprotected'))

            props.viewprotected = False
            self.assertFalse(props.viewprotected)

        # internal data
        self.assertEqual(
            o_getattr(props, '_data'),
            {'unprotected': True, 'viewprotected': False}
        )
        self.assertTrue(o_getattr(props, '_context') is context)
        self.assertEqual(o_getattr(props, '_permissions'), permissions)

        # ``__copy__``
        props_copy = copy.copy(props)
        self.assertFalse(props_copy is props)
        self.assertEqual(props_copy.__class__, ProtectedProperties)
        self.assertEqual(
            o_getattr(props_copy, '_data'),
            o_getattr(props, '_data')
        )
        self.assertTrue(o_getattr(props, '_context') is context)
        self.assertEqual(o_getattr(props, '_permissions'), permissions)

        # ``__deepcopy__``
        props_deepcopy = copy.deepcopy(props)
        self.assertFalse(props_deepcopy is props)
        self.assertEqual(props_deepcopy.__class__, ProtectedProperties)
        self.assertEqual(
            o_getattr(props_deepcopy, '_data'),
            o_getattr(props, '_data')
        )
        self.assertTrue(o_getattr(props, '_context') is context)
        self.assertEqual(o_getattr(props, '_permissions'), permissions)

    def test_XMLProperties(self):
        # There's a convenience object for XML input and output

        # Create temp directory
        tempdir = tempfile.mkdtemp()

        # Create XML properties with path and optional data
        props = XMLProperties(
            os.path.join(tempdir, 'props.xml'),
            data={'foo': u'äöüß'}
        )
        self.assertTrue(IProperties.providedBy(props))

        # Testing helper functions
        self.assertEqual(props._keys(), ['foo'])
        self.assertEqual(props._values(), [u'äöüß'])

        # XML properties can be datetime objects
        props.effective = datetime(2010, 1, 1, 10, 15)
        props.empty = ''

        # XML properties can be multi valued...
        props.keywords = ['a', datetime(2010, 1, 1, 10, 15), '']

        # ...or dict/odict instance
        props.dictlike = odict([('a', 'foo'), ('b', 'bar'), ('c', '')])

        # Nothing added yet
        self.assertEqual(os.listdir(tempdir), [])

        # Call props, file is now written to disk
        props()
        self.assertEqual(os.listdir(tempdir), ['props.xml'])

        # Check file contents
        with open(os.path.join(tempdir, 'props.xml')) as file:
            lines = file.read().split('\n')
        self.assertEqual(lines, [
            '<properties>',
            '  <foo>&#228;&#246;&#252;&#223;</foo>',
            '  <effective>2010-01-01T10:15:00</effective>',
            '  <empty></empty>',
            '  <keywords>',
            '    <item>a</item>',
            '    <item>2010-01-01T10:15:00</item>',
            '    <item></item>',
            '  </keywords>',
            '  <dictlike>',
            '    <elem>',
            '      <key>a</key>',
            '      <value>foo</value>',
            '    </elem>',
            '    <elem>',
            '      <key>b</key>',
            '      <value>bar</value>',
            '    </elem>',
            '    <elem>',
            '      <key>c</key>',
            '      <value></value>',
            '    </elem>',
            '  </dictlike>',
            '</properties>',
            ''
        ])

        # Overwrite ``foo`` and add ``bar`` properties; Note that even markup
        # can be used safely
        props.foo = u'foo'
        props.bar = u'<bar>äöü</bar>'

        # Call props and check result
        props()
        with open(os.path.join(tempdir, 'props.xml')) as file:
            lines = file.read().split('\n')
        self.assertEqual(lines, [
            '<properties>',
            '  <foo>foo</foo>',
            '  <effective>2010-01-01T10:15:00</effective>',
            '  <empty></empty>',
            '  <keywords>',
            '    <item>a</item>',
            '    <item>2010-01-01T10:15:00</item>',
            '    <item></item>',
            '  </keywords>',
            '  <dictlike>',
            '    <elem>',
            '      <key>a</key>',
            '      <value>foo</value>',
            '    </elem>',
            '    <elem>',
            '      <key>b</key>',
            '      <value>bar</value>',
            '    </elem>',
            '    <elem>',
            '      <key>c</key>',
            '      <value></value>',
            '    </elem>',
            '  </dictlike>',
            '  <bar>&lt;bar&gt;&#228;&#246;&#252;&lt;/bar&gt;</bar>',
            '</properties>',
            ''
        ])

        # Create XML properties from existing file
        props = XMLProperties(os.path.join(tempdir, 'props.xml'))
        self.assertEqual(
            props._keys(),
            ['foo', 'effective', 'empty', 'keywords', 'dictlike', 'bar']
        )

        self.assertEqual(
            props._values(),
            [
                u'foo',
                datetime(2010, 1, 1, 10, 15),
                u'',
                [u'a', datetime(2010, 1, 1, 10, 15), u''],
                odict([('a', 'foo'), ('b', 'bar'), ('c', None)]),
                u'<bar>äöü</bar>'
            ]
        )

        # Delete property
        del props['foo']
        self.assertEqual(
            props._keys(),
            ['effective', 'empty', 'keywords', 'dictlike', 'bar']
        )

        err = self.expect_error(
            KeyError,
            lambda: props.__delitem__('inexistent')
        )
        self.assertEqual(
            str(err).strip('u'),
            "'property inexistent does not exist'"
        )

        # Call and check results
        props()
        with open(os.path.join(tempdir, 'props.xml')) as file:
            lines = file.read().split('\n')
        self.assertEqual(lines, [
            '<properties>',
            '  <effective>2010-01-01T10:15:00</effective>',
            '  <empty></empty>',
            '  <keywords>',
            '    <item>a</item>',
            '    <item>2010-01-01T10:15:00</item>',
            '    <item></item>',
            '  </keywords>',
            '  <dictlike>',
            '    <elem>',
            '      <key>a</key>',
            '      <value>foo</value>',
            '    </elem>',
            '    <elem>',
            '      <key>b</key>',
            '      <value>bar</value>',
            '    </elem>',
            '    <elem>',
            '      <key>c</key>',
            '      <value>None</value>',
            '    </elem>',
            '  </dictlike>',
            '  <bar>&lt;bar&gt;&#228;&#246;&#252;&lt;/bar&gt;</bar>',
            '</properties>',
            ''
        ])

        # Change order of odict and check results
        props.dictlike = odict([('b', 'bar'), ('a', 'foo')])
        props()
        with open(os.path.join(tempdir, 'props.xml')) as file:
            lines = file.read().split('\n')
        self.assertEqual(lines, [
            '<properties>',
            '  <effective>2010-01-01T10:15:00</effective>',
            '  <empty></empty>',
            '  <keywords>',
            '    <item>a</item>',
            '    <item>2010-01-01T10:15:00</item>',
            '    <item></item>',
            '  </keywords>',
            '  <dictlike>',
            '    <elem>',
            '      <key>b</key>',
            '      <value>bar</value>',
            '    </elem>',
            '    <elem>',
            '      <key>a</key>',
            '      <value>foo</value>',
            '    </elem>',
            '  </dictlike>',
            '  <bar>&lt;bar&gt;&#228;&#246;&#252;&lt;/bar&gt;</bar>',
            '</properties>',
            ''
        ])

        # internal data
        self.assertTrue(o_getattr(props, '_path').endswith('props.xml'))
        self.assertEqual(o_getattr(props, '_data'), odict([
            ('effective', datetime(2010, 1, 1, 10, 15)),
            ('empty', u''),
            ('keywords', [
                u'a',
                datetime(2010, 1, 1, 10, 15),
                u''
            ]),
            ('dictlike', odict([
                ('b', 'bar'),
                ('a', 'foo')
            ])),
            ('bar', u'<bar>\xe4\xf6\xfc</bar>')
        ]))

        # ``__copy__``
        props_copy = copy.copy(props)
        self.assertFalse(props_copy is props)
        self.assertEqual(props_copy.__class__, XMLProperties)
        self.assertEqual(
            o_getattr(props_copy, '_data'),
            o_getattr(props, '_data')
        )
        self.assertEqual(
            o_getattr(props_copy, '_path'),
            o_getattr(props, '_path')
        )

        # ``__deepcopy__``
        props_deepcopy = copy.deepcopy(props)
        self.assertFalse(props_deepcopy is props)
        self.assertEqual(props_deepcopy.__class__, XMLProperties)
        self.assertEqual(
            o_getattr(props_deepcopy, '_data'),
            o_getattr(props, '_data')
        )
        self.assertEqual(
            o_getattr(props_copy, '_path'),
            o_getattr(props, '_path')
        )

        # Cleanup
        shutil.rmtree(tempdir)

    def test_ConfigProperties(self):
        # A Properties implementation exists for Config files used by python
        # configparser

        # Create temp directory
        tempdir = tempfile.mkdtemp()

        props = ConfigProperties(
            os.path.join(tempdir, 'props.cfg'),
            data={'foo': 1}
        )
        self.assertTrue(IProperties.providedBy(props))

        # Nothing added yet
        self.assertEqual(os.listdir(tempdir), [])

        # Call props, file is now written to disk
        props()
        self.assertEqual(os.listdir(tempdir), ['props.cfg'])

        # Check file contents
        with open(os.path.join(tempdir, 'props.cfg')) as file:
            data = file.read()
        self.assertEqual(data, '[properties]\nfoo = 1\n\n')

        # Overwrite ``foo`` and add ``bar`` properties
        props.foo = u'foo'
        props.bar = u'bar'
        props.baz = u'äöü'

        # Call props and check result
        props()
        with open(os.path.join(tempdir, 'props.cfg'), 'rb') as file:
            lines = file.read().split(b'\n')
        self.assertEqual(lines, [
            b'[properties]',
            b'foo = foo',
            b'bar = bar',
            b'baz = \xc3\xa4\xc3\xb6\xc3\xbc',
            b'',
            b''
        ])

        # Create config properties from existing file
        props = ConfigProperties(os.path.join(tempdir, 'props.cfg'))
        self.assertEqual(props.foo, u'foo')
        self.assertEqual(props.bar, u'bar')
        self.assertEqual(props.baz, u'äöü')

        # Test ``__getitem__``
        self.assertEqual(props['foo'], u'foo')
        self.expect_error(KeyError, lambda: props['inexistent'])

        # Test ``get``
        self.assertEqual(props.get('foo'), u'foo')
        self.assertEqual(props.get('inexistent', u'default'), u'default')

        # Test ``__contains__``
        self.assertTrue('foo' in props)
        self.assertFalse('inexistent' in props)

        # Delete property
        err = self.expect_error(KeyError, lambda: props.__delitem__('inexistent'))
        expected = "'property inexistent does not exist'"
        self.assertEqual(str(err).strip('u'), expected)

        del props['foo']
        self.assertTrue(props.foo is None)

        # Call and check results
        props()
        with open(os.path.join(tempdir, 'props.cfg'), 'rb') as file:
            lines = file.read().split(b'\n')
        self.assertEqual(lines, [
            b'[properties]',
            b'bar = bar',
            b'baz = \xc3\xa4\xc3\xb6\xc3\xbc',
            b'',
            b''
        ])
        del props['baz']

        props = ConfigProperties(os.path.join(tempdir, 'config.cfg'))
        props['foo'] = 'foo'

        # internal data
        self.assertTrue(o_getattr(props, '_path').endswith('config.cfg'))
        self.assertEqual(o_getattr(props, '_data'), {})
        self.assertTrue(o_getattr(props, '_config') is props.config())

        # ``__copy__``
        props_copy = copy.copy(props)
        self.assertFalse(props_copy is props)
        self.assertEqual(props_copy.__class__, ConfigProperties)
        self.assertEqual(
            o_getattr(props_copy, '_data'),
            o_getattr(props, '_data')
        )
        self.assertEqual(
            o_getattr(props_copy, '_path'),
            o_getattr(props, '_path')
        )
        self.assertTrue(isinstance(
            o_getattr(props_copy, '_config'),
            configparser.ConfigParser
        ))
        strio = StringIO()
        props_copy.config().write(strio)
        strio.seek(0)
        self.assertEqual(strio.readlines(), [
            '[properties]\n',
            'foo = foo\n',
            '\n'
        ])

        # ``__deepcopy__``
        props_deepcopy = copy.deepcopy(props)
        self.assertFalse(props_deepcopy is props)
        self.assertEqual(props_deepcopy.__class__, ConfigProperties)
        self.assertEqual(
            o_getattr(props_deepcopy, '_data'),
            o_getattr(props, '_data'),
        )
        self.assertEqual(
            o_getattr(props_deepcopy, '_path'),
            o_getattr(props, '_path')
        )
        self.assertTrue(isinstance(
            o_getattr(props_copy, '_config'),
            configparser.ConfigParser
        ))
        strio = StringIO()
        props_deepcopy.config().write(strio)
        strio.seek(0)
        self.assertEqual(strio.readlines(), [
            '[properties]\n',
            'foo = foo\n',
            '\n'
        ])

        # Cleanup
        shutil.rmtree(tempdir)



from cone.app import compat
from cone.app import testing
from cone.app.browser import set_related_view
from cone.app.browser.contents import ContentsTile
from cone.app.browser.contents import ContentsViewLink
from cone.app.browser.contents import listing
from cone.app.browser.table import TableSlice
from cone.app.browser.utils import make_url
from cone.app.model import BaseNode
from cone.app.model import node_info
from cone.app.testing.mock import CopySupportNode
from cone.app.testing.mock import WorkflowNode
from cone.tile import render_tile
from cone.tile.tests import TileTestCase
from datetime import datetime
from datetime import timedelta
from node.base import BaseNode as NodeBaseNode
from node.behaviors import Order
from plumber import plumbing
from pyramid.exceptions import HTTPForbidden
from pyramid.security import ACLDenied
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Deny
from pyramid.security import Everyone


class NeverShownChild(BaseNode):
    __acl__ = [(Deny, Everyone, ALL_PERMISSIONS)]


class TestBrowserContents(TileTestCase):
    layer = testing.security

    def create_dummy_model(self):
        created = datetime(2011, 3, 14)
        delta = timedelta(1)
        modified = created + delta
        model = BaseNode()
        for i in range(19):
            model[str(i)] = BaseNode()
            model[str(i)].properties.action_view = True
            model[str(i)].properties.action_edit = True
            model[str(i)].properties.action_delete = True
            model[str(i)].metadata.title = str(i) + ' Title'
            model[str(i)].metadata.creator = 'admin ' + str(19 - i)
            model[str(i)].metadata.created = created
            model[str(i)].metadata.modified = modified
            created = created + delta
            modified = modified + delta
        model['nevershown'] = NeverShownChild()
        model['no_app_node'] = NodeBaseNode()
        return model

    def test_sorted_rows(self):
        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')
        contents.model = self.create_dummy_model()
        contents.request = self.layer.new_request()

        # ``sorted_rows`` returns sorted listing items. ``start``, ``end``,
        # ``sort`` and ``order`` are expected by this function
        with self.layer.authenticated('manager'):
            res = contents.sorted_rows(None, None, 'created', 'desc')[0]['title']
            self.checkOutput('...0 Title...', res)

            res = contents.sorted_rows(None, None, 'created', 'desc')[-1]['title']
            self.checkOutput('...18 Title...', res)

            res = contents.sorted_rows(None, None, 'created', 'asc')[0]['title']
            self.checkOutput('...18 Title...', res)

            res = contents.sorted_rows(None, None, 'created', 'asc')[-1]['title']
            self.checkOutput('...0 Title...', res)

    def test_slice(self):
        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')
        contents.model = self.create_dummy_model()
        request = contents.request = self.layer.new_request()

        # ``contents.slice.slice`` return current batch start and positions
        self.assertTrue(isinstance(contents.slice, TableSlice))
        self.assertEqual(contents.slice.slice, (0, 15))

        request.params['b_page'] = '1'
        self.assertEqual(contents.slice.slice, (15, 30))

        del request.params['b_page']

        # ``contents.slice.rows`` return the current sorted row data for listing.
        with self.layer.authenticated('manager'):
            # Items returned by default sorting
            res = contents.slice.rows[0]['title']
            self.checkOutput('...0 Title...', res)

            res = contents.slice.rows[-1]['title']
            self.checkOutput('...14 Title...', res)

            # Inverse order
            request.params['order'] = 'asc'
            res = contents.slice.rows[0]['title']
            self.checkOutput('...18 Title...', res)

            res = contents.slice.rows[-1]['title']
            self.checkOutput('...4 Title...', res)

            # Switch batch page with inversed order
            request.params['b_page'] = '1'
            res = contents.slice.rows[0]['title']
            self.checkOutput('...3 Title...', res)

            res = contents.slice.rows[-1]['title']
            self.checkOutput('...0 Title...', res)

            # Reset order and batch page
            del request.params['order']
            del request.params['b_page']

            # Sort by creator
            request.params['sort'] = 'creator'
            self.assertEqual([row['creator'] for row in contents.slice.rows], [
                'admin 1', 'admin 10', 'admin 11', 'admin 12', 'admin 13',
                'admin 14', 'admin 15', 'admin 16', 'admin 17', 'admin 18',
                'admin 19', 'admin 2', 'admin 3', 'admin 4', 'admin 5'
            ])

            request.params['b_page'] = '1'
            self.assertEqual([row['creator'] for row in contents.slice.rows], [
                'admin 6', 'admin 7', 'admin 8', 'admin 9'
            ])

            # Sort by created
            request.params['b_page'] = '0'
            request.params['sort'] = 'created'

            self.assertEqual(
                contents.slice.rows[0]['created'],
                datetime(2011, 3, 14, 0, 0)
            )
            self.assertEqual(
                contents.slice.rows[-1]['created'],
                datetime(2011, 3, 28, 0, 0)
            )

            request.params['b_page'] = '1'
            request.params['sort'] = 'modified'

            self.assertEqual(
                contents.slice.rows[0]['modified'],
                datetime(2011, 3, 30, 0, 0)
            )

            self.assertEqual(
                contents.slice.rows[-1]['modified'],
                datetime(2011, 4, 2, 0, 0)
            )

            del request.params['b_page']
            del request.params['sort']

    def test_batch(self):
        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')
        contents.model = self.create_dummy_model()
        request = contents.request = self.layer.new_request()

        with self.layer.authenticated('manager'):
            rendered = contents.batch

        expected = (
            '<li class="active">\n          '
            '<a href="javascript:void(0)">1</a>'
        )
        self.assertTrue(rendered.find(expected) != -1)

        expected = (
            'http://example.com/?b_page=1&amp;'
            'order=desc&amp;size=15&amp;sort=created'
        )
        self.assertTrue(rendered.find(expected) != -1)

        # Change page
        request.params['b_page'] = '1'
        with self.layer.authenticated('manager'):
            rendered = contents.batch

        expected = (
            '<li class="active">\n          '
            '<a href="javascript:void(0)">2</a>'
        )
        self.assertTrue(rendered.find(expected) != -1)

        expected = (
            'http://example.com/?b_page=0&amp;'
            'order=desc&amp;size=15&amp;sort=created'
        )
        self.assertTrue(rendered.find(expected) != -1)

        # Change sort and order. Sort is proxied by batch
        request.params['sort'] = 'modified'
        with self.layer.authenticated('manager'):
            rendered = contents.batch

        expected = (
            'http://example.com/?b_page=0&amp;'
            'order=desc&amp;size=15&amp;sort=modified'
        )
        self.assertTrue(rendered.find(expected) != -1)

    def test_authenticated(self):
        model = self.create_dummy_model()
        request = self.layer.new_request()

        # Rendering fails unauthorized, 'list' permission is required
        rule = request.has_permission('list', model)
        self.assertTrue(isinstance(rule, ACLDenied))

        err = self.expectError(
            HTTPForbidden,
            render_tile,
            model,
            request,
            'contents'
        )
        self.checkOutput("""
        Unauthorized: tile
        <cone.app.browser.contents.ContentsTile object at ...> failed
        permission check
        """, str(err))

        # Render authenticated
        with self.layer.authenticated('manager'):
            request.params['sort'] = 'modified'
            request.params['b_page'] = '1'
            rendered = render_tile(model, request, 'contents')
            expected = (
                '<a href="http://example.com/?b_page=1&amp;'
                'order=desc&amp;size=15&amp;sort=title"'
            )
            self.assertTrue(rendered.find(expected) != -1)

    def test_copysupport(self):
        # Copysupport Attributes
        model = CopySupportNode()
        model['child'] = CopySupportNode()
        request = self.layer.new_request()

        with self.layer.authenticated('manager'):
            rendered = render_tile(model, request, 'contents')

        expected = 'class="selectable copysupportitem"'
        self.assertTrue(rendered.find(expected) > -1)

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            cut_url = compat.quote(make_url(request, node=model['child']))
            request.cookies['cone.app.copysupport.cut'] = cut_url
            rendered = render_tile(model, request, 'contents')

        expected = 'class="selectable copysupportitem copysupport_cut"'
        self.assertTrue(rendered.find(expected) > -1)

    def test_ContentsViewLink(self):
        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')

        model = BaseNode(name='root')
        request = self.layer.new_request()

        view_link = contents.view_link
        self.assertTrue(isinstance(contents.view_link, ContentsViewLink))

        with self.layer.authenticated('max'):
            res = view_link(model, request)
        expected = 'ajax:target="http://example.com/root"'
        self.assertTrue(res.find(expected) > -1)

        set_related_view(request, 'someview')
        with self.layer.authenticated('max'):
            res = view_link(model, request)
        expected = 'ajax:target="http://example.com/root?contenttile=someview"'
        self.assertTrue(res.find(expected) > -1)

    def test_workflow_state(self):
        root = BaseNode(name='root')
        root['wf'] = WorkflowNode()
        request = self.layer.new_request()

        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')
        contents.model = root
        contents.request = request

        with self.layer.authenticated('max'):
            res = contents.sorted_rows(
                0,
                1,
                contents.default_sort,
                contents.default_order
            )
        self.assertEqual(res[0].css, ' state-initial')

    @testing.reset_node_info_registry
    def test_node_type(self):
        @node_info(
            name='mynode')
        class MyNode(BaseNode):
            pass

        root = BaseNode()
        root['child'] = MyNode()
        request = self.layer.new_request()

        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')
        contents.model = root
        contents.request = request

        with self.layer.authenticated('max'):
            res = contents.sorted_rows(
                0,
                1,
                contents.default_sort,
                contents.default_order
            )
        self.assertEqual(res[0].css, ' node-type-mynode')

    def test_filtered_children(self):
        root = BaseNode()

        child_1 = root['1'] = BaseNode()
        md_1 = child_1.metadata
        md_1.title = 'Child 1'
        md_1.creator = 'max'

        child_2 = root['2'] = BaseNode()
        md_2 = child_2.metadata
        md_2.title = 'Child 2'
        md_2.creator = 'sepp'

        tmpl = 'cone.app:browser/templates/table.pt'
        contents = ContentsTile(tmpl, None, 'contents')
        contents.model = root
        contents.request = self.layer.new_request()

        res = contents.filtered_children
        self.assertEqual(res, [])

        contents.request = self.layer.new_request()
        with self.layer.authenticated('max'):
            res = contents.filtered_children
        self.assertEqual(len(res), 2)

        request = contents.request = self.layer.new_request()
        request.params['term'] = 'sepp'
        with self.layer.authenticated('max'):
            res = contents.filtered_children
        self.assertEqual(len(res), 1)

        request = contents.request = self.layer.new_request()
        request.params['term'] = 'Child 1'
        with self.layer.authenticated('max'):
            res = contents.filtered_children
        self.assertEqual(len(res), 1)

        request = contents.request = self.layer.new_request()
        request.params['term'] = 'hild'
        with self.layer.authenticated('max'):
            res = contents.filtered_children
        self.assertEqual(len(res), 2)

        request = contents.request = self.layer.new_request()
        request.params['term'] = 'foo'
        with self.layer.authenticated('max'):
            res = contents.filtered_children
        self.assertEqual(len(res), 0)

    def test_listing(self):
        model = self.create_dummy_model()
        request = self.layer.new_request()

        with self.layer.authenticated('max'):
            res = listing(model, request)
        self.assertTrue(res.text.startswith('<!DOCTYPE html>'))

    def test_move_actions(self):
        @plumbing(Order)
        class OrderableNode(BaseNode):
            pass

        node = OrderableNode()
        node.properties.action_move = True
        node['a'] = BaseNode()
        node['b'] = BaseNode()
        node['c'] = BaseNode()

        with self.layer.authenticated('manager'):
            rendered = render_tile(node, self.layer.new_request(), 'contents')

        self.assertEqual(rendered.count('toolbaraction-move-up'), 2)
        self.assertEqual(rendered.count('toolbaraction-move-down'), 2)



from cone.app import compat
from cone.app import testing
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.ajax import AjaxMessage
from cone.app.browser.copysupport import PasteAction
from cone.app.browser.utils import make_url
from cone.app.model import BaseNode
from cone.app.model import node_info
from cone.app.testing.mock import CopySupportNode
from cone.tile.tests import TileTestCase


class TestBrowserCopysupport(TileTestCase):
    layer = testing.security

    @testing.reset_node_info_registry
    def test_copysupport(self):
        @node_info(
            name='copy_support_node_a',
            title='CopySupportNodeA',
            addables=['copy_support_node_a', 'copy_support_node_b'])
        class CopySupportNodeA(CopySupportNode):
            pass

        @node_info(
            name='copy_support_node_b',
            title='CopySupportNodeB',
            addables=['copy_support_node_b'])
        class CopySupportNodeB(CopySupportNode):
            pass

        root = CopySupportNodeA()
        source = root['source'] = CopySupportNodeA()
        source['a_child'] = CopySupportNodeA()
        source['b_child'] = CopySupportNodeB()
        target = root['target'] = CopySupportNodeB()

        request = self.layer.new_request()
        copy_url = compat.quote(make_url(request, node=source['a_child']))
        request.cookies['cone.app.copysupport.copy'] = copy_url

        paste_tile = PasteAction(None, 'render', '')
        paste_tile(target, request)

        self.checkOutput("""
        Pasted 0 items<br /><strong>Pasting of 1 items
        failed</strong><br />Violation. 'CopySupportNodeB' is not allowed
        to contain 'CopySupportNodeA'
        """, request.environ['cone.app.continuation'][0].payload)

        copy_url = compat.quote(make_url(request, node=source['b_child']))
        request.cookies['cone.app.copysupport.copy'] = copy_url
        del request.environ['cone.app.continuation']

        paste_tile(target, request)
        self.assertEqual(target.messages, ['Called: target'])
        target.messages = []

        self.checkOutput("""
        cone.app.copysupport.copy=; Max-Age=0; Path=/; expires=...
        """, request.response.headers['Set-Cookie'])

        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][0],
            AjaxMessage
        ))
        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][1],
            AjaxEvent
        ))
        self.assertEqual(
            request.environ['cone.app.continuation'][1].target,
            'http://example.com/target?contenttile=listing'
        )

        self.checkOutput("""
        <class '...CopySupportNodeA'>: None
          <class '...CopySupportNodeA'>: source
            <class '...CopySupportNodeA'>: a_child
            <class '...CopySupportNodeB'>: b_child
          <class '...CopySupportNodeB'>: target
            <class '...CopySupportNodeB'>: b_child
        """, root.treerepr())

        target.properties.action_paste_tile = 'custom'
        del request.environ['cone.app.continuation']

        paste_tile(target, request)
        self.assertEqual(target.messages, ['Called: target'])
        target.messages = []

        self.assertEqual(
            request.environ['cone.app.continuation'][1].target,
            'http://example.com/target?contenttile=custom'
        )

        self.checkOutput("""
        <class '...CopySupportNodeA'>: None
          <class '...CopySupportNodeA'>: source
            <class '...CopySupportNodeA'>: a_child
            <class '...CopySupportNodeB'>: b_child
          <class '...CopySupportNodeB'>: target
            <class '...CopySupportNodeB'>: b_child
            <class '...CopySupportNodeB'>: b_child-1
        """, root.treerepr())

        cut_url = compat.quote(make_url(request, node=source['b_child']))
        request.cookies['cone.app.copysupport.cut'] = cut_url
        del request.cookies['cone.app.copysupport.copy']
        del request.environ['cone.app.continuation']

        paste_tile(target, request)
        self.assertEqual(target.messages, ['Called: target'])
        self.assertEqual(source.messages, ['Called: source'])
        target.messages = []
        source.messages = []

        self.checkOutput("""
        cone.app.copysupport.cut=; Max-Age=0; Path=/; expires=...
        """, request.response.headers['Set-Cookie'])

        self.checkOutput("""
        <class '...CopySupportNodeA'>: None
          <class '...CopySupportNodeA'>: source
            <class '...CopySupportNodeA'>: a_child
          <class '...CopySupportNodeB'>: target
            <class '...CopySupportNodeB'>: b_child
            <class '...CopySupportNodeB'>: b_child-1
            <class '...CopySupportNodeB'>: b_child-2
        """, root.treerepr())

        cut_url = compat.quote(make_url(request, node=source['a_child']))
        request.cookies['cone.app.copysupport.cut'] = cut_url
        del request.environ['cone.app.continuation']

        paste_tile(target, request)
        self.checkOutput("""
        <class '...CopySupportNodeA'>: None
          <class '...CopySupportNodeA'>: source
            <class '...CopySupportNodeA'>: a_child
          <class '...CopySupportNodeB'>: target
            <class '...CopySupportNodeB'>: b_child
            <class '...CopySupportNodeB'>: b_child-1
            <class '...CopySupportNodeB'>: b_child-2
        """, root.treerepr())

        self.checkOutput("""
        Pasted 0 items<br /><strong>Pasting of 1 items
        failed</strong><br />Violation. 'CopySupportNodeB' is not
        allowed to contain 'CopySupportNodeA'
        """, request.environ['cone.app.continuation'][0].payload)

        cut_url = compat.quote(make_url(request, node=source))
        del request.environ['cone.app.continuation']
        request.cookies['cone.app.copysupport.cut'] = cut_url

        paste_tile(root['source']['a_child'], request)
        self.checkOutput("""
        Pasted 0 items<br /><strong>Pasting of 1 items
        failed</strong><br />Cannot paste cut object to child of it: source
        """, request.environ['cone.app.continuation'][0].payload)

        cut_url = '::'.join([
            compat.quote(make_url(request, node=target['b_child'])),
            compat.quote(make_url(request, node=target['b_child-1'])),
        ])
        request.cookies['cone.app.copysupport.cut'] = cut_url
        del request.environ['cone.app.continuation']

        paste_tile(source, request)
        self.assertEqual(source.messages, ['Called: source'])
        self.assertEqual(target.messages, ['Called: target'])
        source.messages = []
        target.messages = []

        self.checkOutput("""
        <class '...CopySupportNodeA'>: None
          <class '...CopySupportNodeA'>: source
            <class '...CopySupportNodeA'>: a_child
            <class '...CopySupportNodeB'>: b_child
            <class '...CopySupportNodeB'>: b_child-1
          <class '...CopySupportNodeB'>: target
            <class '...CopySupportNodeB'>: b_child-2
        """, root.treerepr())

        root['unknown_source'] = BaseNode()
        root['unknown_target'] = BaseNode()

        cut_url = compat.quote(make_url(request, node=root['unknown_source']))
        request.cookies['cone.app.copysupport.cut'] = cut_url
        del request.environ['cone.app.continuation']

        paste_tile(target, request)
        self.checkOutput("""
        Pasted 0 items<br /><strong>Pasting of 1 items
        failed</strong><br />Cannot paste 'unknown_source'. Unknown source
        """, request.environ['cone.app.continuation'][0].payload)

        cut_url = compat.quote(make_url(request, node=source['b_child']))
        request.cookies['cone.app.copysupport.cut'] = cut_url
        del request.environ['cone.app.continuation']

        paste_tile(root['unknown_target'], request)
        self.checkOutput("""
        Pasted 0 items<br /><strong>Pasting of 1 items
        failed</strong><br />Cannot paste to 'unknown_target'. Unknown target
        """, request.environ['cone.app.continuation'][0].payload)

        del request.cookies['cone.app.copysupport.cut']
        del request.environ['cone.app.continuation']

        paste_tile(root['unknown_target'], request)
        self.assertEqual(
            request.environ['cone.app.continuation'][0].payload,
            u'Nothing to paste'
        )



from cone.app import testing
from cone.app.browser.exception import forbidden_view
from cone.app.browser.exception import internal_server_error
from cone.app.browser.exception import not_found_view
from cone.app.model import BaseNode
from cone.tile import render_tile
from cone.tile.tests import TileTestCase
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import render_view_to_response


class TestBrowserException(TileTestCase):
    layer = testing.security

    def test_internal_server_error(self):
        # When requests are performed, and an uncaught exception is raised,
        # ``internal_server_error`` view is invoked. Response either represents
        # an error page or a JSON response containing bdajax continuation
        # definitions which display the traceback in an error dialog if request
        # was a bdajax action
        request = self.layer.new_request()
        self.checkOutput("""
        200 OK
        Content-Type: text/html; charset=UTF-8
        Content-Length: ...
        <h1>An error occured</h1>
        <p>
          <a href="/">HOME</a>
          <hr />
        </p>
        <pre>Traceback (most recent call last):
          None...
        </pre>
        """, str(internal_server_error(request)))

        request = self.layer.new_request(xhr=1)
        res = str(internal_server_error(request))
        self.assertTrue(res.find('200 OK') > -1)
        self.assertTrue(res.find('Content-Type: application/json') > -1)
        self.assertTrue(res.find('"continuation"') > -1)
        self.assertTrue(res.find('"type": "message"') > -1)
        expected = '"payload": "<pre>Traceback (most recent call last):'
        self.assertTrue(res.find(expected) > -1)
        self.assertTrue(res.find('"selector": null') > -1)
        self.assertTrue(res.find('"payload": ""') > -1)
        self.assertTrue(res.find('"mode": "NONE"') > -1)
        self.assertTrue(res.find('"selector": "NONE"') > -1)

        with self.layer.authenticated('admin'):
            request = self.layer.new_request()

            self.checkOutput("""
            200 OK
            Content-Type: text/html; charset=UTF-8
            Content-Length: ...
            <h1>An error occured</h1>
            <p>
              <a href="/">HOME</a>
              <hr />
            </p>
            <pre>Traceback (most recent call last):
              None...
            </pre>
            """, str(internal_server_error(request)))

            request = self.layer.new_request(xhr=1)
            res = str(internal_server_error(request))

        self.assertTrue(res.find('200 OK') > -1)
        self.assertTrue(res.find('Content-Type: application/json') > -1)
        self.assertTrue(res.find('"continuation"') > -1)
        self.assertTrue(res.find('"type": "message"') > -1)
        expected = '"payload": "<pre>Traceback (most recent call last):'
        self.assertTrue(res.find(expected) > -1)
        self.assertTrue(res.find('"selector": null') > -1)
        self.assertTrue(res.find('"payload": ""') > -1)
        self.assertTrue(res.find('"mode": "NONE"') > -1)
        self.assertTrue(res.find('"selector": "NONE"') > -1)

    def test_forbidden(self):
        # Forbidden tile
        root = BaseNode()
        model = root['model'] = BaseNode()
        request = self.layer.new_request()

        self.checkOutput("""
        \n\n  <div>\n    <h1>Unauthorized</h1>\n
        <p>You are not allowed to access this resource.</p>\n  </div>\n\n\n
        """, render_tile(model, request, 'unauthorized'))

        # Forbidden view. Unauthenticated renders login form.
        context = HTTPForbidden()
        request = self.layer.new_request()
        request.context = BaseNode()
        res = render_view_to_response(context, request=request).text
        self.assertTrue(res.find('id="input-loginform-login"') > -1)

        # Authenticated renders unauthorized as content tile.
        with self.layer.authenticated('admin'):
            res = render_view_to_response(context, request=request).text
            self.assertTrue(res.find('<h1>Unauthorized</h1>') > -1)

    def test_json_forbidden(self):
        context = HTTPForbidden()
        request = self.layer.new_request(type='json')
        request.context = BaseNode()
        res = render_view_to_response(context, request=request)
        self.assertEqual(res.text, '{}')
        self.assertEqual(res.status, '403 Forbidden')

    def test_not_found(self):
        # Not Found tile
        root = BaseNode()
        model = root['model'] = BaseNode()
        request = self.layer.new_request()

        self.checkOutput("""
        \n\n  <div>\n    <h1>Not Found</h1>\n
        <p>The requested resource cannot be found.</p>\n  </div>\n\n\n
        """, render_tile(model, request, 'not_found'))

        # Not Found view. Always renders not found as content tile.
        context = HTTPNotFound()
        request = self.layer.new_request()
        request.context = BaseNode()
        res = render_view_to_response(context, request=request).text
        self.assertTrue(res.find('<h1>Not Found</h1>') > -1)

        with self.layer.authenticated('admin'):
            res = render_view_to_response(context, request=request).text
            self.assertTrue(res.find('<h1>Not Found</h1>') > -1)

    def test_json_not_found(self):
        context = HTTPNotFound()
        request = self.layer.new_request(type='json')
        request.context = BaseNode()
        res = render_view_to_response(context, request=request)
        self.assertEqual(res.text, '{}')
        self.assertEqual(res.status, '404 Not Found')



from cone.app import testing
from cone.app.browser.actions import ActionContext
from cone.app.browser.content import content_view_action
from cone.app.browser.content import content_view_tile
from cone.app.browser.content import ContentViewAction
from cone.app.browser.contextmenu import context_menu
from cone.app.model import BaseNode
from cone.tile import render_tile
from cone.tile import Tile
from cone.tile.tests import TileTestCase
from zope.interface import implementer
from zope.interface import Interface


class TestBrowserLayout(TileTestCase):
    layer = testing.security

    def test_content_view_tile(self):
        class Model(BaseNode):
            pass

        with self.layer.hook_tile_reg():
            @content_view_tile(
                name='model_tile',
                interface=Model,
                permission='view')
            class ModelContentTile(Tile):
                def render(self):
                    return '<div>Model Content</div>'

        model = Model()
        request = self.layer.new_request()
        with self.layer.authenticated('manager'):
            res = render_tile(model, request, 'model_tile')
        self.assertEqual(res, '<div>Model Content</div>')

        from cone.app.tests.test_browser_content import model_tile_content_view
        self.assertEqual(model_tile_content_view.__doc__, (
            'Dynamically created by '
            'cone.app.browser.content.content_view_tile'
        ))

        with self.layer.authenticated('manager'):
            res = model_tile_content_view(model, request)
        self.assertTrue(res.body.startswith(b'<!DOCTYPE html>') > -1)
        self.assertTrue(res.body.find(b'<div>Model Content</div>') > -1)

    def test_content_view_action(self):
        class Model(BaseNode):
            pass

        @content_view_action(
            name='model_action',
            tilename='model_tile',
            interface=Model,
            permission='view',
            text='Model Action',
            icon='glyphicons glyphicons-star')
        class ModelContentTile(Tile):
            pass

        self.assertTrue('model_action' in context_menu['contentviews'])

        action = context_menu['contentviews']['model_action']
        self.assertIsInstance(action, ContentViewAction)
        self.assertEqual(action.name, 'model_tile')
        self.assertEqual(action.interface, Model)
        self.assertEqual(action.permission, 'view')
        self.assertEqual(action.text, 'Model Action')
        self.assertEqual(action.icon, 'glyphicons glyphicons-star')

        del context_menu['contentviews']['model_action']

    def test_ContentViewAction(self):
        model = BaseNode(name='model')
        request = self.layer.new_request()

        action = ContentViewAction(name='content_action')
        action.model = model
        action.request = request
        self.assertEqual(action.href, 'http://example.com/model/content_action')
        self.assertTrue(action.display)

        ActionContext(model, request, 'content_action')
        self.assertTrue(action.selected)

        del request.environ['action_context']
        self.assertFalse(action.selected)

        action = ContentViewAction(name='content_action', permission='view')
        action.model = model
        action.request = request
        self.assertFalse(action.display)
        with self.layer.authenticated('manager'):
            self.assertTrue(action.display)

        action = ContentViewAction(name='content_action', interface=BaseNode)
        action.model = model
        action.request = request
        self.assertTrue(action.display)

        class InvalidClassContext(object):
            pass

        action = ContentViewAction(
            name='content_action',
            interface=InvalidClassContext
        )
        action.model = model
        action.request = request
        self.assertFalse(action.display)

        class IContextInterface(Interface):
            pass

        action = ContentViewAction(
            name='content_action',
            interface=IContextInterface
        )
        action.model = model
        action.request = request
        self.assertFalse(action.display)

        @implementer(IContextInterface)
        class InterfaceModel(BaseNode):
            pass

        action.model = InterfaceModel()
        self.assertTrue(action.display)



from cone.app import testing
from cone.app.browser.ajax import AjaxAction
from cone.app.browser.form import Form
from cone.app.browser.form import ProtectedAttributesForm
from cone.app.browser.form import YAMLForm
from cone.app.model import BaseNode
from cone.tile import render_tile
from cone.tile import tile
from cone.tile.tests import TileTestCase
from plumber import plumbing
from pyramid.security import ACLAllowed
from pyramid.security import ACLDenied
from webob.exc import HTTPFound
from yafowil.base import factory


class TestBrowserForm(TileTestCase):
    layer = testing.security

    def test_Form(self):
        formtile = Form(None, None, 'plainform')
        err = self.expectError(
            NotImplementedError,
            formtile.prepare
        )
        expected = '``prepare`` function must be provided by deriving object.'
        self.assertEqual(str(err), expected)

        subscriptions = []

        with self.layer.hook_tile_reg():
            @tile(name='subscriptionform')
            class SubscriptionForm(Form):
                ajax = False
                show = False
                # test flags
                next_as_redirect = True
                continuation_as_list = False

                def prepare(self):
                    form = factory(
                        u'form',
                        name='subscriptionform',
                        props={
                            'action': self.nodeurl,
                            'class': 'foo',
                            'class_add': 'bar'
                        })
                    form['email'] = factory(
                        'field:label:error:text',
                        props={
                            'required': 'No email given',
                            'label': 'E-Mail',
                        })
                    form['subscribe'] = factory(
                        'submit',
                        props={
                            'action': 'subscribe',
                            'expression': True,
                            'handler': self.subscribe,
                            'next': self.next,
                            'label': 'Subscribe',
                        })
                    self.form = form

                def subscribe(self, widget, data):
                    # Do subscription here
                    subscriptions.append('subscribe on "%s"' % self.model.name)

                def next(self, request):
                    url = 'http://example.com'
                    if self.ajax_request:
                        # return as single value, gets list on request.environ
                        cont = AjaxAction(url, 'content', 'inner', '#content')
                        if self.continuation_as_list:
                            return [cont]
                        else:
                            return cont
                    if self.next_as_redirect:
                        return HTTPFound(url)
                    # return anything else to be rendered
                    return '<div>success!</div>'

        model = BaseNode()
        model.__name__ = 'dummymodel'
        request = self.layer.new_request()

        # Render form. ``form.show`` returns false, render empty string
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')
            self.assertEqual(rendered, u'')

        # Set show to True
        SubscriptionForm.show = True
        # Render form. no action is triggered and no input is given
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        expected = 'class="bar foo"'
        self.assertTrue(rendered.find(expected) != -1)

        expected = 'action="http://example.com/dummymodel"'
        self.assertTrue(rendered.find(expected) != -1)

        expected = 'id="form-subscriptionform"'
        self.assertTrue(rendered.find(expected) != -1)

        expected = 'name="subscriptionform.email"'
        self.assertTrue(rendered.find(expected) != -1)

        expected = 'name="action.subscriptionform.subscribe"'
        self.assertTrue(rendered.find(expected) != -1)

        # Trigger subscribe action and set empty email value. Results in a form
        # with error message since email is required
        request = self.layer.new_request()
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = ''

        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        expected = 'No email given'
        self.assertTrue(rendered.find(expected) != -1)

        # Trigger subscribe action and set valid email value. Now the action
        # handler and next handler are triggered
        request = self.layer.new_request()
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = 'john.doe@example.com'

        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        self.assertEqual(rendered, '')
        self.assertEqual(subscriptions, ['subscribe on "dummymodel"'])
        subscriptions = []

        # The form was rendered as non ajax form, so we expect an HTTPFound
        # instance on request
        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))

        # Even if we commit as ajax form, it is treaten as normal form since
        # ajax flag is set to False (defaults to True)
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = 'john.doe@example.com'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        self.assertEqual(rendered, '')
        self.assertEqual(subscriptions, ['subscribe on "dummymodel"'])
        subscriptions = []

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))

        # We can return markup insted of HTTPFound if we want to render inplace
        # instead of redirection
        SubscriptionForm.next_as_redirect = False

        request = self.layer.new_request()
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = 'john.doe@example.com'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        self.assertEqual(rendered, '<div>success!</div>')
        self.assertEqual(subscriptions, ['subscribe on "dummymodel"'])
        subscriptions = []

        self.assertFalse('redirect' in request.environ)

        SubscriptionForm.next_as_redirect = True

        # Try with ajax True.
        SubscriptionForm.ajax = True

        request = self.layer.new_request()
        request.params['ajax'] = '1'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        expected = 'class="ajax bar foo"'
        self.assertTrue(rendered.find(expected) != -1)

        # If submitted without ajax flag on request, still get HTTPFound
        request = self.layer.new_request()
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = 'john.doe@example.com'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        self.assertEqual(rendered, '')
        self.assertEqual(subscriptions, ['subscribe on "dummymodel"'])
        subscriptions = []

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))

        # Submit with ajax flag
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = 'john.doe@example.com'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        self.assertEqual(rendered, '')
        self.assertEqual(subscriptions, ['subscribe on "dummymodel"'])
        subscriptions = []

        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][0],
            AjaxAction
        ))

        # Ajax continuation may be returned as list
        SubscriptionForm.continuation_as_list = True

        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['action.subscriptionform.subscribe'] = '1'
        request.params['subscriptionform.email'] = 'john.doe@example.com'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'subscriptionform')

        self.assertEqual(rendered, '')
        self.assertEqual(subscriptions, ['subscribe on "dummymodel"'])
        subscriptions = []

        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][0],
            AjaxAction
        ))

    def test_YAMLForm(self):
        with self.layer.hook_tile_reg():
            @tile(name='yamlsubscriptionform')
            @plumbing(YAMLForm)
            class YAMLSubscriptionForm(Form):
                action_resource = 'yamlsubscriptionform'
                form_template = 'cone.app.testing:dummy_form.yaml'

        model = BaseNode()
        model.__name__ = 'dummymodel'
        request = self.layer.new_request()

        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'yamlsubscriptionform')

        expected = (
            'action="http://example.com/dummymodel/yamlsubscriptionform"'
        )
        self.assertTrue(rendered.find(expected) > -1)

        # Instead of ``form_template`` attribute, ``form_template_path`` can be
        # used for backward compatibility
        with self.layer.hook_tile_reg():
            @tile(name='yamlsubscriptionform2')
            class YAMLSubscriptionForm2(YAMLSubscriptionForm):
                action_resource = 'yamlsubscriptionform2'
                form_template = None
                form_template_path = 'cone.app.testing:dummy_form.yaml'

        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'yamlsubscriptionform2')

        expected = (
            'action="http://example.com/dummymodel/yamlsubscriptionform2"'
        )
        self.assertTrue(rendered.find(expected) > -1)

        # form flavor add renders form action URL on parent
        root = BaseNode(name='root')
        model = root['child'] = BaseNode()

        YAMLSubscriptionForm.form_flavor = 'add'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'yamlsubscriptionform')
        expected = (
            'action="http://example.com/root/yamlsubscriptionform"'
        )
        self.assertTrue(rendered.find(expected) > -1)

    def test_ProtectedAttributesForm(self):
        # ProtectedAttributesForm plumbing behavior
        with self.layer.hook_tile_reg():
            @tile(name='protectedattributesform')
            @plumbing(ProtectedAttributesForm)
            class MyProtectedAttributesForm(Form):
                attribute_permissions = {
                    'protectedfield': ('manage', 'edit')
                }

                def prepare(self):
                    form = factory(
                        u'form',
                        name='protectedattributesform',
                        props={
                            'action': self.nodeurl,
                        })
                    form['protectedfield'] = factory(
                        u'field:label:text',
                        value=u'Protectedfield',
                        mode=self.mode_for('protectedfield')
                    )
                    self.form = form

        model = BaseNode()
        model.__name__ = 'dummymodel'
        request = self.layer.new_request()

        with self.layer.authenticated('viewer'):
            rule = request.has_permission('edit', model)
            self.assertTrue(isinstance(rule, ACLDenied))
            rendered = render_tile(model, request, 'protectedattributesform')

        self.checkOutput("""
        <form
        action="http://example.com/dummymodel"
        class="ajax"
        enctype="multipart/form-data"
        id="form-protectedattributesform"
        method="post"
        novalidate="novalidate"></form>
        """, rendered)

        with self.layer.authenticated('editor'):
            rule = request.has_permission('edit', model)
            self.assertTrue(isinstance(rule, ACLAllowed))
            rendered = render_tile(model, request, 'protectedattributesform')

        self.checkOutput("""
        <form ...<div class="display-text"
        id="display-protectedattributesform-protectedfield">Protectedfield</div></div></form>
        """, rendered)

        with self.layer.authenticated('manager'):
            rule = request.has_permission('manage', model)
            self.assertTrue(isinstance(rule, ACLAllowed))
            rendered = render_tile(model, request, 'protectedattributesform')

        self.checkOutput("""
        <form ...<input class="text"
        id="input-protectedattributesform-protectedfield"
        name="protectedattributesform.protectedfield"
        type="text" value="Protectedfield" /></div></form>
        """, rendered)

        # Test default attribute permissions
        MyProtectedAttributesForm.attribute_permissions = dict()
        self.assertEqual(
            MyProtectedAttributesForm.attribute_default_permissions,
            ('edit', 'view')
        )

        with self.layer.authenticated('viewer'):
            rule = request.has_permission('view', model)
            self.assertTrue(isinstance(rule, ACLAllowed))
            rendered = render_tile(model, request, 'protectedattributesform')

        self.checkOutput("""
        <form ...<div class="display-text"
        id="display-protectedattributesform-protectedfield">Protectedfield</div></div></form>
        """, rendered)

        with self.layer.authenticated('editor'):
            rule = request.has_permission('edit', model)
            self.assertTrue(isinstance(rule, ACLAllowed))
            rendered = render_tile(model, request, 'protectedattributesform')

        self.checkOutput("""
        <form ...<input class="text"
        id="input-protectedattributesform-protectedfield"
        name="protectedattributesform.protectedfield"
        type="text" value="Protectedfield" /></div></form>
        """, rendered)



from cone.app import compat
from cone.app.browser import RelatedViewConsumer
from cone.app.browser.batch import Batch
from cone.app.browser.utils import format_date
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.utils import node_path
from cone.tile import Tile
from node.utils import safe_decode
from plumber import plumbing


class RowData(dict):

    def __init__(self, selectable=False, target=None, css=''):
        self.selectable = selectable
        self.target = target
        self.css = css


@plumbing(RelatedViewConsumer)
class Table(Tile):
    """Abstract table tile. Provides rendering of sortable, batched tables.

    A subclass of this tile must be registered under the same name as defined
    at ``self.table_tile_name``, normally bound to template
    ``cone.app:browser/templates/table.pt``
    """
    wrapper_binding = 'batchclicked sortclicked'
    table_id = 'table'
    table_css = ''
    table_tile_name = 'table'
    col_defs = []
    default_sort = None
    default_order = None
    default_slicesize = 15
    query_whitelist = []
    show_title = True
    show_filter = False
    show_slicesize = True
    head_additional = None
    display_table_header = True
    display_table_footer = True
    ajax_path = None
    ajax_path_event = None

    table_length_size = 'col-xs-4 col-sm3'
    table_filter_size = 'col-xs-3'

    @property
    def slice(self):
        return TableSlice(self, self.model, self.request)

    @property
    def batch(self):
        return TableBatch(self)(self.model, self.request)

    @property
    def slicesize(self):
        return int(self.request.params.get('size', self.default_slicesize))

    @property
    def slicesizes(self):
        return [i * self.default_slicesize for i in range(1, 5)]

    @property
    def table_title(self):
        return self.model.metadata.title

    @property
    def slice_target(self):
        return self.make_url({
            'sort': self.sort_column,
            'order': self.sort_order,
            'term': self.filter_term,
        })

    @property
    def filter_target(self):
        return self.make_url({
            'sort': self.sort_column,
            'order': self.sort_order,
            'size': self.slicesize,
        })

    @property
    def filter_term(self):
        term = self.request.params.get('term')
        if term:
            term = term.encode('utf-8') if compat.IS_PY2 else term
            term = compat.unquote(term)
            term = term.decode('utf-8') if compat.IS_PY2 else term
        return term

    @property
    def sort_column(self):
        return self.request.params.get('sort', self.default_sort)

    @property
    def sort_order(self):
        return self.request.params.get('order', self.default_order)

    @property
    def sort_index(self):
        """Index of recent sort column.
        """
        col = self.sort_column
        idx = 0
        for col_def in self.col_defs:
            key = col_def.get('sort_key')
            if key == col:
                return idx
            idx += 1

    def make_query(self, params):
        """Create query considering ``query_whitelist``.

        :param params: Dictionary with query parameters.
        :return: Query as string.
        """
        p = dict()
        for param in self.query_whitelist:
            p[param] = self.request.params.get(param, '')
        p.update(params)
        return make_query(**p)

    def make_url(self, params, path=None, include_view=False):
        """Create URL considering ``query_whitelist``.

        :param params: Dictionary with query parameters.
        :param path: Optional model path, if ``None``, path gets taken from
            ``self.model``
        :param include_view: Boolean whether to include
            ``self.related_view`` to URL.
        :return: URL as string.
        """
        return safe_decode(make_url(
            self.request,
            path=path,
            node=None if path else self.model,
            resource=self.related_view if include_view else None,
            query=self.make_query(params)))

    def format_date(self, dt):
        return format_date(dt)

    def th_defs(self, sortkey):
        cur_sort = self.sort_column
        cur_order = self.sort_order
        selected = cur_sort == sortkey
        alter = selected and cur_order == 'desc'
        order = alter and 'asc' or 'desc'
        params = {
            'b_page': self.request.params.get('b_page', '0'),
            'sort': sortkey,
            'order': order,
            'size': self.slicesize,
            'term': self.filter_term,
        }
        url = self.make_url(params)
        css = selected and order or ''
        return css, url

    @property
    def item_count(self):
        raise NotImplementedError("Abstract table does not implement "
                                  "``item_count``.")

    def sorted_rows(self, start, end, sort, order):
        raise NotImplementedError("Abstract table does not implement "
                                  "``sorted_rows``.")


class TableSlice(object):

    def __init__(self, table_tile, model, request):
        self.table_tile = table_tile
        self.model = model
        self.request = request

    @property
    def slice(self):
        current = int(self.request.params.get('b_page', '0'))
        start = current * self.table_tile.slicesize
        end = start + self.table_tile.slicesize
        return start, end

    @property
    def rows(self):
        start, end = self.slice
        return self.table_tile.sorted_rows(
            start, end,
            self.table_tile.sort_column,
            self.table_tile.sort_order)


class TableBatch(Batch):

    def __init__(self, table_tile):
        self.table_tile = table_tile
        self.name = table_tile.table_id + 'batch'
        self.related_view = table_tile.related_view
        self.ajax_path = self.ajax_path
        self.ajax_path_event = self.ajax_path_event

    @property
    def display(self):
        return len(self.vocab) > 1

    @property
    def vocab(self):
        ret = list()
        path = node_path(self.model)
        count = self.table_tile.item_count
        slicesize = self.table_tile.slicesize
        pages = count // slicesize
        if count % slicesize != 0:
            pages += 1
        current = self.request.params.get('b_page', '0')
        params = {
            'sort': self.table_tile.sort_column,
            'order': self.table_tile.sort_order,
            'size': slicesize,
            'term': self.table_tile.filter_term,
        }
        for term in self.table_tile.query_whitelist:
            params[term] = self.request.params.get(term, '')
        for i in range(pages):
            params['b_page'] = str(i)
            query = make_query(**params)
            url = make_url(
                self.request,
                path=path,
                # resource=self.related_view,
                query=query
            )
            ret.append({
                'page': '%i' % (i + 1),
                'current': current == str(i),
                'visible': True,
                'url': url,
            })
        return ret



from cone.app import testing
from cone.app.model import BaseNode
from cone.app.model import Translation
from node.utils import UNSET
from plumber import plumbing
from yafowil.base import factory
from yafowil.tests import fxml
from yafowil.tests import NodeTestCase


@plumbing(Translation)
class TranslationNode(BaseNode):
    child_constraints = None


class TestBrowserTranslation(NodeTestCase):
    layer = testing.security

    def test_translation_extractor(self):
        widget = factory(
            'translation:text',
            name='field',
            props={
                'label': 'Field'
            })

        request = self.layer.new_request()
        data = widget.extract(request)
        self.assertEqual(data.extracted, UNSET)

        request.params['field.en'] = u'Value EN'
        request.params['field.de'] = u'Value DE'
        data = widget.extract(request)
        self.assertIsInstance(data.extracted, dict)
        self.assertEqual(data.extracted, {
            'en': u'Value EN',
            'de': u'Value DE'
        })

    def test_translation_extractor_factory(self):
        widget = factory(
            'translation:text',
            name='field',
            props={
                'label': 'Field',
                'factory': TranslationNode
            })

        request = self.layer.new_request()
        data = widget.extract(request)
        self.assertEqual(data.extracted, UNSET)

        request.params['field.en'] = u'Value EN'
        request.params['field.de'] = u'Value DE'
        data = widget.extract(request)
        self.assertIsInstance(data.extracted, TranslationNode)
        self.assertEqual(data.extracted.items(), [
            ('en', u'Value EN'),
            ('de', u'Value DE')
        ])

    def test_extraction_preset_value(self):
        widget = factory(
            'translation:text',
            name='field',
            value={
                'en': u'Value EN',
                'de': u'Value DE'
            },
            props={
                'label': 'Field'
            })

        request = self.layer.new_request()
        data = widget.extract(request)
        self.assertEqual(data.extracted, UNSET)

        request.params['field.en'] = u'Value EN new'
        request.params['field.de'] = u'Value DE new'
        data = widget.extract(request)
        self.assertEqual(data.extracted, {
            'en': u'Value EN new',
            'de': u'Value DE new'
        })

    def test_extraction_preset_value_factory(self):
        value = TranslationNode()
        value['en'] = u'Value EN'
        value['de'] = u'Value DE'
        widget = factory(
            'translation:text',
            name='field',
            value=value,
            props={
                'label': 'Field',
                'factory': TranslationNode
            })

        request = self.layer.new_request()
        data = widget.extract(request)
        self.assertEqual(data.extracted, UNSET)

        request.params['field.en'] = u'Value EN new'
        request.params['field.de'] = u'Value DE new'
        data = widget.extract(request)
        self.assertIsInstance(data.extracted, TranslationNode)
        self.assertEqual(data.extracted.items(), [
            ('en', u'Value EN new'),
            ('de', u'Value DE new')
        ])

    def test_translation_edit_renderer_no_value(self):
        widget = factory(
            'field:translation:text',
            'field',
            props={
                'label': 'Field'
            })

        self.checkOutput("""
        <div class="field" id="field-field">
          <ul class="nav nav-pills translation-nav">
            <li class="active">
              <a href="#translation-field-en">EN</a>
            </li>
            <li>
              <a href="#translation-field-de">DE</a>
            </li>
          </ul>
          <div class="translation-fields">
            <div id="translation-field-en">
              <input class="text" id="input-field-en" name="field.en"
                     type="text" value=""/>
            </div>
            <div id="translation-field-de">
              <input class="text" id="input-field-de" name="field.de"
                     type="text" value=""/>
            </div>
          </div>
        </div>
        """, fxml(widget()))

    def test_translation_edit_renderer_preset_value(self):
        widget = factory(
            'field:translation:text',
            'field',
            value={
                'en': u'Value EN',
                'de': u'Value DE'
            },
            props={
                'label': 'Field'
            })

        self.checkOutput("""
        ...
        <div id="translation-field-en">
          <input class="text" id="input-field-en" name="field.en"
                 type="text" value="Value EN"/>
        </div>
        <div id="translation-field-de">
          <input class="text" id="input-field-de" name="field.de"
                 type="text" value="Value DE"/>
        </div>
        ...
        """, fxml(widget()))

    def test_translation_edit_renderer_preset_translation_value(self):
        value = TranslationNode()
        value['en'] = u'Value EN'
        value['de'] = u'Value DE'
        widget = factory(
            'field:translation:text',
            'field',
            value=value,
            props={
                'label': 'Field',
                'factory': TranslationNode
            })

        self.checkOutput("""
        ...
        <div id="translation-field-en">
          <input class="text" id="input-field-en" name="field.en"
                 type="text" value="Value EN"/>
        </div>
        <div id="translation-field-de">
          <input class="text" id="input-field-de" name="field.de"
                 type="text" value="Value DE"/>
        </div>
        ...
        """, fxml(widget()))

    def test_render_after_extraction_no_preset_value(self):
        widget = factory(
            'field:translation:text',
            'field',
            props={
                'label': 'Field'
            })

        request = self.layer.new_request()
        request.params['field.en'] = u'Value EN'
        request.params['field.de'] = u'Value DE'
        data = widget.extract(request)
        self.checkOutput("""
        ...
        <div id="translation-field-en">
          <input class="text" id="input-field-en" name="field.en"
                 type="text" value="Value EN"/>
        </div>
        <div id="translation-field-de">
          <input class="text" id="input-field-de" name="field.de"
                 type="text" value="Value DE"/>
        </div>
        ...
        """, fxml(widget(data=data)))

    def test_render_after_extraction_preset_value(self):
        widget = factory(
            'field:translation:text',
            'field',
            value={
                'en': u'Value EN',
                'de': u'Value DE'
            },
            props={
                'label': 'Field'
            })

        request = self.layer.new_request()
        request.params['field.en'] = u'Value EN new'
        request.params['field.de'] = u'Value DE new'
        data = widget.extract(request)
        self.checkOutput("""
        ...
        <div id="translation-field-en">
          <input class="text" id="input-field-en" name="field.en"
                 type="text" value="Value EN new"/>
        </div>
        <div id="translation-field-de">
          <input class="text" id="input-field-de" name="field.de"
                 type="text" value="Value DE new"/>
        </div>
        ...
        """, fxml(widget(data=data)))

    def test_render_after_extraction_preset_translation_value(self):
        value = TranslationNode()
        value['en'] = u'Value EN'
        value['de'] = u'Value DE'
        widget = factory(
            'field:translation:text',
            'field',
            value=value,
            props={
                'label': 'Field',
                'factory': TranslationNode
            })

        request = self.layer.new_request()
        request.params['field.en'] = u'Value EN new'
        request.params['field.de'] = u'Value DE new'
        data = widget.extract(request)
        self.checkOutput("""
        ...
        <div id="translation-field-en">
          <input class="text" id="input-field-en" name="field.en"
               type="text" value="Value EN new"/>
        </div>
        <div id="translation-field-de">
          <input class="text" id="input-field-de" name="field.de"
               type="text" value="Value DE new"/>
        </div>
        ...
        """, fxml(widget(data=data)))

    def test_required(self):
        widget = factory(
            'field:error:translation:text',
            'field',
            value={
                'en': u'Value EN',
                'de': u'Value DE'
            },
            props={
                'label': 'Field',
                'required': 'Field is mandatory'
            })

        request = self.layer.new_request()
        request.params['field.en'] = u''
        request.params['field.de'] = u'Value DE'
        data = widget.extract(request)
        self.checkOutput("""
        <div class="field" id="field-field">
          <div class="error">
            <div class="errormessage">Field is mandatory</div>
            <ul class="nav nav-pills translation-nav">
              <li class="active error">
                <a href="#translation-field-en">* EN</a>
              </li>
              <li>
                <a href="#translation-field-de">DE</a>
              </li>
            </ul>
            <div class="translation-fields">
              <div id="translation-field-en">
                <input class="required text" id="input-field-en" name="field.en"
                       required="required" type="text" value=""/>
              </div>
              <div id="translation-field-de">
                <input class="required text" id="input-field-de" name="field.de"
                       required="required" type="text" value="Value DE"/>
              </div>
            </div>
          </div>
        </div>
        """, fxml(widget(data=data)))

    def test_translation_display_renderer(self):
        widget = factory(
            'field:translation:text',
            'field',
            value={
                'en': u'Value EN',
                'de': u'Value DE'
            },
            props={
                'label': 'Field'
            },
            mode='display')

        self.checkOutput("""
        <div class="field" id="field-field">
          <ul class="nav nav-pills translation-nav">
            <li class="active">
              <a href="#translation-field-en">EN</a>
            </li>
            <li>
              <a href="#translation-field-de">DE</a>
            </li>
          </ul>
          <div class="translation-fields">
            <div id="translation-field-en">
              <div class="display-text" id="display-field-en">Value EN</div>
            </div>
            <div id="translation-field-de">
              <div class="display-text" id="display-field-de">Value DE</div>
            </div>
          </div>
        </div>
        """, fxml(widget()))



from cone.app.interfaces import ILayout
from cone.app.model import AppNode
from cone.app.model import BaseNode
from cone.app.model import CopySupport
from cone.app.model import Layout
from cone.app.model import Properties
from cone.app.security import DEFAULT_ACL
from cone.app.security import PrincipalACL
from cone.app.workflow import WorkflowACL
from cone.app.workflow import WorkflowState
from node.behaviors import Attributes
from node.behaviors import DefaultInit
from node.behaviors import MappingAdopt
from node.behaviors import MappingNode
from node.behaviors import OdictStorage
from node.utils import instance_property
from plumber import plumbing
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Allow
from pyramid.security import Deny
from pyramid.security import Everyone
from pyramid.static import static_view
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


static_resources = static_view('static', use_subpath=True)


@plumbing(WorkflowState, WorkflowACL)
class WorkflowNode(BaseNode):
    workflow_name = u'dummy'
    workflow_tsf = None

    @property
    def properties(self):
        props = Properties()
        props.in_navtree = True
        return props

    def __call__(self):
        pass


class InexistentWorkflowNode(WorkflowNode):
    workflow_name = u'inexistent'
    workflow_tsf = None


class StateACLWorkflowNode(WorkflowNode):
    state_acls = {
        'initial': [
            (Allow, 'role:manager', ['manage', 'edit', 'change_state']),
            (Allow, Everyone, ['login']),
            (Deny, Everyone, ALL_PERMISSIONS),
        ],
        'final': [
            (Allow, 'role:manager', ['view', 'edit', 'change_state']),
            (Deny, Everyone, ALL_PERMISSIONS),
        ],
    }


class IWorkflowNode(Interface):
    pass


@plumbing(WorkflowState)
@implementer(IWorkflowNode)
class InterfaceWorkflowNode(BaseNode):
    workflow_name = u'dummy'
    workflow_tsf = None


@plumbing(
    PrincipalACL,
    AppNode,
    MappingAdopt,
    Attributes,
    DefaultInit,
    MappingNode,
    OdictStorage)
class SharingNode(object):

    @property
    def __acl__(self):
        return DEFAULT_ACL

    @instance_property
    def principal_roles(self):
        return dict()


@plumbing(CopySupport)
class CopySupportNode(BaseNode):

    def __init__(self, name=None, parent=None):
        super(CopySupportNode, self).__init__(name=name, parent=parent)
        self.messages = []

    def __call__(self):
        self.messages.append('Called: {}'.format(self.name))


###############################################################################
# B/C
###############################################################################

@implementer(ILayout)
@adapter(Interface)
def default_layout(context):
    config = Layout()
    config.mainmenu = True
    config.mainmenu_fluid = False
    config.livesearch = True
    config.personaltools = True
    config.columns_fluid = False
    config.pathbar = True
    config.sidebar_left = ['navtree']
    config.sidebar_left_grid_width = 3
    config.content_grid_width = 9
    return config


class LayoutConfigNode(BaseNode):

    @property
    def layout(self):
        return default_layout(self)



from cone.app import ApplicationNodeTraverser
from cone.app import DefaultLayoutConfig
from cone.app import get_root
from cone.app import layout_config
from cone.app import main_hook
from cone.app import make_remote_addr_middleware
from cone.app import testing
from cone.app.interfaces import ILayoutConfig
from cone.app.model import BaseNode
from cone.app.model import LayoutConfig
from node.base import BaseNode as NodeBaseNode
from node.tests import NodeTestCase
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.router import Router
from pyramid.static import static_view
from pyramid.testing import DummyRequest
from yafowil import resources
from yafowil.base import factory as yafowil_factory
import cone.app


def tmp_root_node(fn):
    def wrapper(*a):
        cone.app.root = cone.app.default_root_node_factory({})
        try:
            fn(*a)
        finally:
            cone.app.root = None
    return wrapper


class TestApp(NodeTestCase):

    @tmp_root_node
    def test_get_root(self):
        root = cone.app.get_root()
        self.assertTrue(str(root).startswith("<AppRoot object 'None' at"))

        # AppRoot contains a settings node by default
        self.assertTrue('settings' in root.factories.keys())

        # Settings contains metadata.title by default
        self.assertEqual(list(root['settings'].metadata.keys()), ['title'])
        self.assertEqual(root['settings'].metadata.title, u'settings')

        # Settings is displayed in navtree by default
        self.assertEqual(
            sorted(root['settings'].properties.keys()),
            ['icon', 'in_navtree', 'skip_mainmenu']
        )
        self.assertFalse(root['settings'].properties.in_navtree)
        self.assertTrue(root['settings'].properties.skip_mainmenu)

    @tmp_root_node
    def test_register_plugin(self):
        cone.app.register_plugin('dummy', BaseNode)

        root = cone.app.get_root()
        self.assertTrue('dummy' in root.factories.keys())

        err = self.expect_error(
            ValueError,
            lambda: cone.app.register_plugin('dummy', BaseNode)
        )
        expected = "Entry with name 'dummy' already registered."
        self.assertEqual(str(err), expected)

    @tmp_root_node
    def test_register_plugin_config(self):
        cone.app.register_plugin_config('dummy', BaseNode)

        root = cone.app.get_root()
        self.assertTrue('dummy' in root['settings'].factories.keys())

        err = self.expect_error(
            ValueError,
            lambda: cone.app.register_plugin_config('dummy', BaseNode)
        )
        expected = "Config with name 'dummy' already registered."
        self.assertEqual(str(err), expected)

    def test_main(self):
        # main hook
        hooks = dict(called=0)

        def custom_main_hook(configurator, global_config, settings):
            hooks['called'] += 1

        cone.app.register_main_hook(custom_main_hook)

        @main_hook
        def decorated_main_hook(configurator, global_config, settings):
            hooks['called'] += 1

        # set auth tkt factory``
        factory = cone.app.auth_tkt_factory(secret='12345')
        self.assertTrue(isinstance(factory, AuthTktAuthenticationPolicy))

        # ACL factory
        factory = cone.app.acl_factory()
        self.assertTrue(isinstance(factory, ACLAuthorizationPolicy))

        # yafowil resources
        def dummy_get_plugin_names(ns=None):
            return ['yafowil.addon']

        get_plugin_names_origin = resources.get_plugin_names
        resources.get_plugin_names = dummy_get_plugin_names

        yafowil_addon_name = 'yafowil.addon'
        js = [{
            'group': 'yafowil.addon.common',
            'resource': 'widget.js',
            'order': 20,
        }]
        css = [{
            'group': 'yafowil.addon.common',
            'resource': 'widget.css',
            'order': 20,
        }]
        yafowil_factory.register_theme(
            'default',
            yafowil_addon_name,
            'yafowil_addon_resources',
            js=js,
            css=css
        )

        # remember original main template
        main_template_orgin = cone.app.cfg.main_template

        # settings
        settings = {
            'cone.admin_user': 'admin',
            'cone.admin_password': 'admin',
            'cone.auth_secret': '12345',
            'cone.auth_reissue_time': '300',
            'cone.auth_max_age': '600',
            'cone.main_template': 'package.browser:templates/main.pt',
            # ensure custom root node factory gets invoked
            'cone.root.node_factory': 'cone.app.default_root_node_factory',
            # ensure dummy main hooks called
            'cone.plugins': 'cone.app.tests'
        }

        # main
        router = cone.app.main({}, **settings)
        self.assertTrue(isinstance(router, Router))
        self.assertEqual(hooks['called'], 2)

        # Remove custom main hook after testing
        cone.app.main_hooks.remove(custom_main_hook)
        cone.app.main_hooks.remove(decorated_main_hook)

        # Check main template was set properly
        self.assertEqual(
            cone.app.cfg.main_template,
            'package.browser:templates/main.pt'
        )

        # reset main template
        cone.app.cfg.main_template = main_template_orgin

        # Check created yafowil addon static view
        self.assertTrue(isinstance(
            cone.app.yafowil_addon_resources,
            static_view
        ))

        # Remove dummy yafowil theme, reset get_plugin_names patch
        # and delete created yafowil addon static view
        resources.get_plugin_names = get_plugin_names_origin
        del yafowil_factory._themes['default'][yafowil_addon_name]
        del cone.app.yafowil_addon_resources

    def test_remote_addr_middleware(self):
        # Remote address middleware
        class DummyApp(object):
            remote_addr = None

            def __call__(self, environ, start_response):
                self.remote_addr = environ['REMOTE_ADDR']

        app = DummyApp()
        middleware = make_remote_addr_middleware(app, {})

        environ = {}
        environ['HTTP_X_REAL_IP'] = '1.2.3.4'
        middleware(environ, None)
        self.assertEqual(app.remote_addr, '1.2.3.4')

    def test_layout_config(self):
        config = layout_config.lookup(model=BaseNode(), request=DummyRequest())
        self.assertIsInstance(config, LayoutConfig)
        self.assertIsInstance(config, DefaultLayoutConfig)
        self.assertTrue(ILayoutConfig.providedBy(config))

        self.assertTrue(config.mainmenu)
        self.assertFalse(config.mainmenu_fluid)
        self.assertTrue(config.livesearch)
        self.assertTrue(config.personaltools)
        self.assertFalse(config.columns_fluid)
        self.assertTrue(config.pathbar)
        self.assertEqual(config.sidebar_left, ['navtree'])
        self.assertEqual(config.sidebar_left_grid_width, 3)
        self.assertEqual(config.content_grid_width, 9)

        @layout_config(BaseNode)
        class BaseNodeLayout(LayoutConfig):
            pass

        class CustomNode1(BaseNode):
            pass

        class CustomNode2(BaseNode):
            pass

        @layout_config(CustomNode1, CustomNode2)
        class CustomNodeLayout(LayoutConfig):
            pass

        config = layout_config.lookup(model=BaseNode(), request=DummyRequest())
        self.assertIsInstance(config, BaseNodeLayout)

        config = layout_config.lookup(model=CustomNode1(), request=DummyRequest())
        self.assertIsInstance(config, CustomNodeLayout)

        config = layout_config.lookup(model=CustomNode2(), request=DummyRequest())
        self.assertIsInstance(config, CustomNodeLayout)

        del layout_config._registry[BaseNode]
        del layout_config._registry[CustomNode1]
        del layout_config._registry[CustomNode2]


class TestTraversal(NodeTestCase):
    layer = testing.security

    def test_ApplicationNodeTraverser(self):
        root = BaseNode()
        root.child_constraints = None
        root['appnode_child'] = BaseNode()
        root['node_child'] = NodeBaseNode()
        root['non_node_child'] = {}

        traverser = ApplicationNodeTraverser(root)
        request = self.layer.new_request()

        request.matchdict['traverse'] = '/view'
        result = traverser(request)
        self.assertTrue(result['context'] is root)
        self.assertEqual(result['view_name'], 'view')
        self.assertEqual(result['traversed'], ())

        request.matchdict['traverse'] = '/appnode_child/view'
        result = traverser(request)
        self.assertTrue(result['context'] is root['appnode_child'])
        self.assertEqual(result['view_name'], 'view')
        self.assertEqual(result['traversed'], ('appnode_child',))

        request.matchdict['traverse'] = '/node_child'
        result = traverser(request)
        self.assertTrue(result['context'] is root)
        self.assertEqual(result['view_name'], 'node_child')
        self.assertEqual(result['traversed'], ())

        request.matchdict['traverse'] = '/non_node_child'
        result = traverser(request)
        self.assertTrue(result['context'] is get_root())
        self.assertEqual(result['view_name'], '')
        self.assertEqual(result['traversed'], ())



from cone.app import compat
from cone.app import get_root
from cone.app.browser.actions import LinkAction
from cone.app.browser.actions import Toolbar
from cone.app.browser.contents import ContentsTile
from cone.app.browser.layout import PathBar
from cone.app.browser.table import RowData
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.browser.utils import request_property
from cone.app.interfaces import INavigationLeaf
from cone.tile import render_tile
from cone.tile import Tile
from cone.tile import tile
from node.interfaces import IUUID
from node.utils import instance_property
from node.utils import LocationIterator
from node.utils import node_by_path
from pyramid.i18n import TranslationStringFactory
from pyramid.threadlocal import get_current_request
from yafowil.base import factory
from yafowil.base import UNSET
from yafowil.common import generic_required_extractor
from yafowil.common import select_display_renderer
from yafowil.common import select_edit_renderer
from yafowil.common import select_extractor
from yafowil.utils import attr_value
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops
from yafowil.utils import Tag
from yafowil.utils import vocabulary


_ = TranslationStringFactory('cone.app')


# XXX: define i18n translation callback for app
tag = Tag(lambda x: x)


def make_refbrowser_query(request, **kw):
    params = {
        'root': request.params['root'],
        'referencable': request.params['referencable'],
        'selected': request.params['selected'],
        'navigable': request.params['navigable'],
    }
    params.update(kw)
    return make_query(**params)


class ReferenceBrowserModelMixin(object):

    @request_property
    def referencable_root(self):
        return node_by_path(
            self.model.root,
            self.request.params['root'].strip('/')
        )

    @request_property
    def referencebrowser_model(self):
        root = self.referencable_root
        if root in LocationIterator(self.model):
            return self.model
        return root


@tile(name='referencebrowser',
      path='templates/referencebrowser.pt',
      permission='view')
class ReferenceBrowser(Tile, ReferenceBrowserModelMixin):

    @property
    def referencebrowser_pathbar(self):
        return render_tile(
            self.referencebrowser_model,
            self.request,
            'referencebrowser_pathbar'
        )

    @property
    def referencelisting(self):
        return render_tile(
            self.referencebrowser_model,
            self.request,
            'referencelisting'
        )


@tile(name='referencebrowser_pathbar',
      path='templates/referencebrowser_pathbar.pt',
      permission='view')
class ReferenceBrowserPathBar(PathBar, ReferenceBrowserModelMixin):

    @property
    def items(self):
        return self.items_for(
            self.referencebrowser_model,
            breakpoint=self.referencable_root
        )

    def item_url(self, node):
        query = make_refbrowser_query(self.request)
        return make_url(self.request, node=node, query=query)

    def item_target(self, node):
        query = make_refbrowser_query(
            self.request, contenttile=node.properties.default_content_tile)
        return make_url(self.request, node=node, query=query)


class ReferenceAction(LinkAction):
    target = None
    href = LinkAction.target

    @property
    def selected_uids(self):
        return [it for it in self.request.params['selected'].split(',') if it]

    @property
    def id(self):
        return 'ref-{}'.format(self.model.uuid)

    @property
    def display(self):
        if not IUUID.providedBy(self.model):
            return False
        referencable = self.request.params['referencable']
        if not referencable:
            return True
        referencable = [it for it in referencable.split(',') if it]
        return self.model.node_info_name in referencable

    def render(self):
        rendered = LinkAction.render(self)
        attrs = {
            'class_': 'reftitle',
            'style': 'display:none;',
        }
        title = self.model.metadata.get('title', self.model.name)
        return rendered + tag('span', title, **attrs)


class ActionAddReference(ReferenceAction):
    css = 'addreference'
    title = _('add_reference', default='Add reference')
    icon = 'glyphicons glyphicons-plus-sign'

    @property
    def enabled(self):
        if IUUID.providedBy(self.model):
            return not str(self.model.uuid) in self.selected_uids
        return False


class ActionRemoveReference(ReferenceAction):
    css = 'removereference'
    title = _('remove_reference', default='Remove reference')
    icon = 'glyphicons glyphicons-minus-sign'

    @property
    def enabled(self):
        if IUUID.providedBy(self.model):
            return str(self.model.uuid) in self.selected_uids
        return False


class ReferencableChildrenLink(LinkAction):
    event = 'contextchanged:.refbrowsersensitiv'

    def __init__(self, table_tile_name, table_id):
        self.table_tile_name = table_tile_name
        self.table_id = table_id

    @property
    def target(self):
        return '{}{}'.format(
            super(ReferencableChildrenLink, self).target,
            make_refbrowser_query(self.request))

    @property
    def text(self):
        return self.model.metadata.get('title', self.model.name)

    @property
    def action(self):
        return '{}:#{}:replace'.format(self.table_tile_name, self.table_id)

    @property
    def display(self):
        return self.permitted('view')

    @property
    def icon(self):
        return self.model.nodeinfo.icon

    def render(self):
        if INavigationLeaf.providedBy(self.model):
            return u'{}&nbsp;{}'.format(
                tag('span', class_=self.icon),
                tag('span', self.text)
            )
        return LinkAction.render(self)


@tile(name='referencelisting', path='templates/table.pt', permission='view')
class ReferenceListing(ContentsTile):
    table_id = 'referencebrowser'
    table_tile_name = 'referencelisting'
    col_defs = [
        {
            'id': 'actions',
            'title': _('table_actions', default='Actions'),
            'sort_key': None,
            'sort_title': None,
            'content': 'structure',
        },
        {
            'id': 'title',
            'title': _('title', default='Title'),
            'sort_key': None,
            'sort_title': None,
            'content': 'structure',
        },
    ]
    query_whitelist = ['root', 'referencable', 'selected', 'navigable']

    @instance_property
    def row_actions(self):
        row_actions = Toolbar()
        row_actions['add'] = ActionAddReference()
        row_actions['remove'] = ActionRemoveReference()
        return row_actions

    @instance_property
    def referencable_children_link(self):
        return ReferencableChildrenLink(self.table_tile_name, self.table_id)

    @property
    def navigable_types(self):
        navigable = self.request.params.get('navigable')
        if not navigable:
            return []
        return navigable.split(',')

    def sorted_rows(self, start, end, sort, order):
        children = self.sorted_children(sort, order)
        rows = list()
        for child in children[start:end]:
            row_data = RowData()
            row_data['actions'] = self.row_actions(child, self.request)
            row_data['title'] = \
                self.referencable_children_link(child, self.request)
            rows.append(row_data)
        return rows

    def sorted_children(self, sort, order):
        children = super(ReferenceListing, self).sorted_children(sort, order)
        navigable_types = self.navigable_types
        if not navigable_types:
            return children
        return [
            child for child in children
            if child.node_info_name in navigable_types
        ]


def fetch_reference_value(widget, data):
    if data.extracted is not UNSET:
        return data.extracted
    if data.value is not UNSET:
        if widget.attrs.get('multivalued'):
            return data.value
        else:
            return data.value[0]
    return attr_value('default', widget, data)


@managedprops('multivalued')
def reference_extractor(widget, data):
    if widget.attrs.get('multivalued'):
        return select_extractor(widget, data)
    if widget.dottedpath not in data.request:
        return UNSET
    return data.request['{}.uid'.format(widget.dottedpath)]


def wrap_ajax_target(rendered, widget, data):
    target = widget.attrs['target']
    if not target:
        request = data.request.request
        request = request if request else get_current_request()
        target = make_url(request, node=get_root())
    if callable(target):
        target = target(widget, data)
    referencable = widget.attrs['referencable']
    if callable(referencable):
        referencable = referencable(widget, data)
    if type(referencable) in compat.ITER_TYPES:
        referencable = ','.join(referencable)
    if not referencable:
        referencable = ''
    navigable = widget.attrs['navigable']
    if callable(navigable):
        navigable = navigable(widget, data)
    if type(navigable) in compat.ITER_TYPES:
        navigable = ','.join(navigable)
    if not navigable:
        navigable = ''
    root = widget.attrs['root']
    if callable(root):
        root = root(widget, data)
    value = fetch_reference_value(widget, data)
    selected = ''
    if widget.attrs['multivalued'] and value:
        selected = ','.join(value)
    elif value:
        selected = value
    query = make_query(**{
        'root': root,
        'referencable': referencable,
        'selected': selected,
        'navigable': navigable,
    })
    target = '{}{}'.format(target, query)
    attrs = {
        'ajax:target': target,
    }
    return tag('span', rendered, **attrs)


def reference_trigger_renderer(widget, data):
    attrs = {
        'class': 'referencebrowser_trigger',
        'data-reference-name': widget.dottedpath
    }
    return data.tag(
        'span',
        tag('i', '', class_='ion-android-share'),
        _('browse', default='Browse'),
        **attrs
    )


def multivalued_reference_vocab(widget, data):
    vocab = list()
    bc_vocab = attr_value('bc_vocabulary', widget, data)
    if bc_vocab:
        for uuid_, label in vocabulary(bc_vocab):
            vocab.append((uuid_, label))
        return vocab
    if widget.dottedpath in data.request:
        value = data.request[widget.dottedpath]
        if isinstance(value, compat.STR_TYPE):
            value = [value]
    else:
        if callable(widget.getter):
            value = widget.getter(widget, data)
        else:
            value = widget.getter
    if not value:
        value = list()
    lookup = widget.attrs.get('lookup')
    for uuid_ in value:
        vocab.append((uuid_, lookup(uuid_) if lookup else uuid_))
    return vocab


def prepare_vocab_property(widget, data):
    if 'vocabulary' in widget.attrs and 'bc_vocabulary' not in widget.attrs:
        widget.attrs['bc_vocabulary'] = widget.attrs['vocabulary']
        del widget.attrs['vocabulary']
    if 'vocabulary' not in widget.attrs:
        widget.attrs['vocabulary'] = multivalued_reference_vocab(widget, data)


def fetch_reference_label(widget, data):
    label = ''
    if widget.dottedpath in data.request:
        label = data.request[widget.dottedpath]
    elif data.value is not UNSET:
        label = data.value[1]
    return label


@managedprops(
    'multivalued', 'vocabulary', 'target',
    'root', 'referencable', 'lookup', 'navigable')
def reference_edit_renderer(widget, data):
    if widget.attrs.get('multivalued'):
        prepare_vocab_property(widget, data)
        rendered = select_edit_renderer(widget, data)
        trigger = reference_trigger_renderer(widget, data)
        return wrap_ajax_target(rendered + trigger, widget, data)
    label = fetch_reference_label(widget, data)
    value = fetch_reference_value(widget, data)
    text_attrs = {
        'type': 'text',
        'value': label,
        'name_': widget.dottedpath,
        'id': cssid(widget, 'input'),
        'class_': cssclasses(widget, data),
        'readonly': 'readonly',
    }
    hidden_attrs = {
        'type': 'hidden',
        'value': value if value else '',
        'name_': '{}.uid'.format(widget.dottedpath),
    }
    rendered = tag('input', **text_attrs) + tag('input', **hidden_attrs)
    trigger = reference_trigger_renderer(widget, data)
    return wrap_ajax_target(rendered + trigger, widget, data)


def reference_display_renderer(widget, data):
    if widget.attrs.get('multivalued'):
        prepare_vocab_property(widget, data)
        return select_display_renderer(widget, data)
    label = fetch_reference_label(widget, data)
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{}'.format(widget.attrs['class'] or 'generic')
    }
    return data.tag('div', label, **attrs)


factory.register(
    'reference',
    extractors=[
        reference_extractor,
        generic_required_extractor
    ],
    edit_renderers=[reference_edit_renderer],
    display_renderers=[reference_display_renderer])

factory.defaults['reference.required_class'] = 'required'

factory.defaults['reference.default'] = ''

factory.defaults['reference.format'] = 'block'

factory.defaults['reference.class'] = 'referencebrowser form-control'

factory.defaults['reference.target'] = None
factory.doc['props']['reference.target'] = """\
Ajax target for reference browser triggering. If not defined, application root
is used.
"""

factory.defaults['reference.root'] = '/'
factory.doc['props']['reference.root'] = """\
Path of reference browser root. Defaults to '/'
"""

factory.defaults['reference.referencable'] = None
factory.doc['props']['reference.referencable'] = """\
Node info name or list of node info names which are referencable. Defaults to
None which means all objects are referenceable, given they implement
``node.interfaces.IUUID`` and provide a node info.
"""

factory.defaults['reference.navigable'] = None
factory.doc['props']['reference.navigable'] = """\
Node info name or list of node info names which are navigable. Defaults to
None which means all objects are navigable, given they implement provide a
node info.
"""

factory.defaults['reference.multivalued'] = False
factory.doc['props']['reference.multivalued'] = """\
Flag whether reference field is multivalued.
"""

factory.defaults['reference.lookup'] = None
factory.doc['props']['reference.lookup'] = """\
Callback accepting reference uid as argument. It is used to lookup the label of
referenced items if multivalued.
"""

factory.doc['props']['reference.vocabulary'] = """\
This property is deprecated and only kept for B/C reasons. Use ``lookup``
instead. If multivalued, provide a vocabulary mapping uids to node names.
"""



from cone.app import security
from cone.app.browser import RelatedViewProvider
from cone.app.browser import render_main_template
from cone.app.browser.ajax import ajax_message
from cone.app.browser.table import RowData
from cone.app.browser.table import Table
from cone.app.ugm import ugm_backend
from cone.tile import Tile
from cone.tile import tile
from plumber import plumbing
from pyramid.i18n import get_localizer
from pyramid.i18n import TranslationStringFactory
from pyramid.view import view_config
from yafowil.base import factory
import logging


logger = logging.getLogger('cone.app')
_ = TranslationStringFactory('cone.app')


@tile(name='sharing',
      path='templates/sharing.pt',
      permission='manage_permissions')
@plumbing(RelatedViewProvider)
class SharingTile(Tile):
    """Tile rendering the sharing table.
    """
    related_view = 'sharing'


@view_config(name='sharing', permission='manage_permissions')
def sharing(model, request):
    """Sharing view.
    """
    return render_main_template(model, request, 'sharing')


@tile(name='local_acl',
      path='templates/table.pt',
      permission='manage_permissions')
class SharingTable(Table):
    table_id = 'localacltable'
    table_tile_name = 'local_acl'
    default_sort = 'principal'
    default_order = 'asc'
    show_filter = True

    @property
    def col_defs(self):
        col_defs = [{
            'id': 'principal',
            'title': _('principal', default='Principal'),
            'sort_key': 'principal',
            'sort_title': _('sort_by_principal',
                            default='Sort by principal'),
            'content': 'string'
        }]
        for role in security.DEFAULT_ROLES:
            col_defs.append({
                'id': role[0],
                'title': role[1],
                'sort_key': None,
                'sort_title': None,
                'content': 'structure',
            })
        return col_defs

    @property
    def table_title(self):
        localizer = get_localizer(self.request)
        title = localizer.translate(self.model.metadata.title)
        return _('sharing_table_title',
                 default='Sharing: ${title}',
                 mapping={'title': title})

    @property
    def item_count(self):
        term = self.filter_term
        if term:
            principals = security.search_for_principals('*%s*' % term)
            return len(principals)
        return len(self.model.principal_roles.keys())

    def sorted_rows(self, start, end, sort, order):
        rows = list()
        term = self.filter_term
        model = self.model
        principal_roles = model.principal_roles
        inheritance = model.role_inheritance
        if term:
            principal_ids = security.search_for_principals('*%s*' % term)
        else:
            if inheritance:
                principal_ids = model.aggregated_roles.keys()
            else:
                principal_ids = principal_roles.keys()
        # XXX: currently always sorted by principal id. Fix to sort by
        #      principal title, needs some refactoring though
        ids = sorted(principal_ids)
        if order == 'desc':
            ids.reverse()
        user_display_attr = ugm_backend.user_display_attr
        group_display_attr = ugm_backend.group_display_attr
        for principal_id in ids[start:end]:
            principal = security.principal_by_id(principal_id)
            if not principal:
                logger.warning('principal %s not found' % principal_id)
                continue
            else:
                if principal_id.startswith('group:'):
                    title = principal.attrs.get(group_display_attr)
                    title = title if title else principal_id[6:]
                else:
                    title = principal.attrs.get(user_display_attr)
                    title = title if title else principal_id
            row_data = RowData()
            row_data['principal'] = title
            ugm_roles = principal.roles
            local_roles = principal_roles.get(principal_id, list())
            if inheritance:
                for role in model.aggregated_roles_for(principal_id):
                    if role not in local_roles:
                        ugm_roles.append(role)
            for role in security.DEFAULT_ROLES:
                inherited = role[0] in ugm_roles
                local = role[0] in local_roles
                row_data[role[0]] = \
                    self._role_column(principal_id, role[0], local, inherited)
            rows.append(row_data)
        return rows

    def _role_column(self, id, role, local, inherited):
        props = {
            'class': 'add_remove_role_for_principal',
            'disabled': inherited,
            'format': 'string',
        }
        props['checked'] = local or inherited
        cb = factory('checkbox', name=id, value=role, props=props)
        ret = '<span ajax:target="%s">' % self.nodeurl
        ret = '%s%s</span>' % (ret, cb())
        return ret


@tile(name='add_principal_role', permission='manage_permissions')
class AddPrincipalRole(Tile):

    def render(self):
        model = self.model
        request = self.request
        try:
            principal_id = request.params['id']
            role = request.params['role']
            roles = model.principal_roles
            if principal_id not in roles:
                model.principal_roles[principal_id] = [role]
                return u''
            existing = set(model.principal_roles[principal_id])
            existing.add(role)
            model.principal_roles[principal_id] = list(existing)
        except Exception as e:
            logger.error(e)
            localizer = get_localizer(self.request)
            message = localizer.translate(
                _(
                    'cannot_add_role_for_principal',
                    default="Can not add role '${role}' for principal '${pid}'"
                ),
                mapping={'role': role, 'pid': principal_id}
            )
            ajax_message(self.request, message, 'error')
        return u''


@tile(name='remove_principal_role', permission='manage_permissions')
class RemovePrincipalRole(Tile):

    def render(self):
        model = self.model
        request = self.request
        try:
            principal_id = request.params['id']
            role = request.params['role']
            roles = model.principal_roles
            if principal_id not in roles:
                raise
            existing = model.principal_roles[principal_id]
            existing.remove(role)
            if not existing:
                del model.principal_roles[principal_id]
            else:
                model.principal_roles[principal_id] = existing
        except Exception as e:
            logger.error(e)
            localizer = get_localizer(self.request)
            message = localizer.translate(
                _(
                    'cannot_remove_role_for_principal',
                    default=(
                        "Can not remove role '${role}' for "
                        "principal '${pid}'"
                    )
                ),
                mapping={'role': role, 'pid': principal_id}
            )
            ajax_message(self.request, message, 'error')
        return u''



from cone.app import testing
from cone.app.browser.actions import _ActionMove
from cone.app.browser.actions import Action
from cone.app.browser.actions import ActionAdd
from cone.app.browser.actions import ActionContext
from cone.app.browser.actions import ActionCopy
from cone.app.browser.actions import ActionCut
from cone.app.browser.actions import ActionDelete
from cone.app.browser.actions import ActionDeleteChildren
from cone.app.browser.actions import ActionEdit
from cone.app.browser.actions import ActionList
from cone.app.browser.actions import ActionMoveDown
from cone.app.browser.actions import ActionMoveUp
from cone.app.browser.actions import ActionPaste
from cone.app.browser.actions import ActionSharing
from cone.app.browser.actions import ActionState
from cone.app.browser.actions import ActionUp
from cone.app.browser.actions import ActionView
from cone.app.browser.actions import ButtonAction
from cone.app.browser.actions import DropdownAction
from cone.app.browser.actions import get_action_context
from cone.app.browser.actions import LinkAction
from cone.app.browser.actions import TemplateAction
from cone.app.browser.actions import TileAction
from cone.app.browser.actions import Toolbar
from cone.app.browser.actions import ViewLink
from cone.app.interfaces import ICopySupport
from cone.app.interfaces import IPrincipalACL
from cone.app.interfaces import IWorkflowState
from cone.app.model import BaseNode
from cone.app.model import NodeInfo
from cone.app.model import Properties
from cone.app.model import register_node_info
from cone.app.testing.mock import CopySupportNode
from cone.app.testing.mock import SharingNode
from cone.app.testing.mock import WorkflowNode
from cone.tile import tile
from cone.tile import Tile
from cone.tile.tests import TileTestCase
from node.behaviors import Order
from plumber import plumbing
from pyramid.security import ACLAllowed
from pyramid.security import ACLDenied


class TestBrowserActions(TileTestCase):
    layer = testing.security

    def test_ActionContext(self):
        model = BaseNode()
        request = self.layer.new_request()

        ac = ActionContext(model, request, 'tile')
        self.assertTrue(request.environ['action_context'] is ac)
        self.assertTrue(get_action_context(request) is ac)

        self.assertEqual(ac.scope, 'tile')

        request.params['bdajax.action'] = 'ajaxaction'
        self.assertEqual(ac.scope, 'ajaxaction')

        request.params['bdajax.action'] = 'layout'
        self.assertEqual(ac.scope, 'content')

        request.params['contenttile'] = 'contenttile'
        self.assertEqual(ac.scope, 'contenttile')

        class PropNode(BaseNode):
            properties = None

        node = PropNode()
        node.properties = Properties()
        node.properties.default_child = 'a'
        node['a'] = PropNode()
        node['a'].properties = Properties()
        node['a'].properties.default_content_tile = 'default'

        request = self.layer.new_request()
        ac = ActionContext(node, request, 'content')
        self.assertEqual(ac.scope, 'default')

    def test_Action_and_Toolbar(self):
        model = BaseNode()
        request = self.layer.new_request()

        err = self.expectError(
            NotImplementedError,
            Action(),
            model,
            request
        )
        self.assertEqual(
            str(err),
            'Abstract ``Action`` does not implement render.'
        )

        result = TileAction()(model, request)
        self.checkOutput("""
        Tile with name '' not found:...
        """, result)

        err = self.expectError(
            ValueError,
            TemplateAction(),
            model,
            request
        )
        self.assertEqual(str(err), 'Relative path not supported: ')

        class DummyAction(Action):
            def render(self):
                return '<a href="">dummy action</a>'

        self.assertEqual(
            DummyAction()(model, request),
            '<a href="">dummy action</a>'
        )

        class DummyTemplateAction(TemplateAction):
            template = u'cone.app.testing:dummy_action.pt'

        self.assertEqual(
            DummyTemplateAction()(model, request),
            u'<a href="">dummy template action</a>'
        )

        with self.layer.hook_tile_reg():
            @tile(name='dummy_action_tile',
                  path='cone.app.testing:dummy_action.pt')
            class DummyActionTile(Tile):
                pass

        class DummyTileAction(TileAction):
            tile = u'dummy_action_tile'

        with self.layer.authenticated('viewer'):
            self.assertEqual(
                DummyTileAction()(model, request),
                u'<a href="">dummy template action</a>'
            )

    def test_Toolbar(self):
        model = BaseNode()
        request = self.layer.new_request()

        class DummyAction(Action):
            def render(self):
                return '<a href="">dummy action</a>'

        tb = Toolbar()
        tb['a'] = DummyAction()

        with self.layer.authenticated('viewer'):
            self.assertEqual(
                tb(model, request),
                u'<div><a href="">dummy action</a></div>'
            )

            tb.css = 'someclass'
            self.assertEqual(
                tb(model, request),
                u'<div class="someclass"><a href="">dummy action</a></div>'
            )

            tb.display = False
            self.assertEqual(tb(model, request), u'')

    def test_abstract_dropdown(self):
        model = BaseNode()
        request = self.layer.new_request()

        err = self.expectError(
            NotImplementedError,
            DropdownAction(),
            model,
            request
        )
        self.checkOutput("""
        ...Abstract ``DropdownAction`` does not implement  ``items``...
        """, str(err))

    def test_LinkAction(self):
        model = BaseNode()
        request = self.layer.new_request()

        rendered = LinkAction()(model, request)
        self.checkOutput("""
        <a
         href="#"
         data-toggle="tooltip"
         data-placement="top"
         ajax:bind="click"
         ajax:target="http://example.com/"
        ></a>
        """, rendered)

        action = LinkAction()
        action.id = 'link_id'
        action.href = 'http://example.com/foo'
        action.css = 'link_action'
        action.title = 'Foo'
        action.action = 'actionname:#content:replace'
        action.event = 'contextchanged:.contextsensitiv'
        action.confirm = 'Do you want to perform?'
        action.overlay = 'someaction'
        action.path = '/foo'
        action.path_target = 'target'
        action.path_action = action.action
        action.path_event = action.event
        action.path_overlay = action.overlay
        action.text = 'Foo'
        rendered = action(model, request)
        self.checkOutput("""
        <a
         id="link_id"
         href="http://example.com/foo"
         class="link_action"
         title="Foo"
         ajax:bind="click"
         ajax:target="http://example.com/"
         ajax:event="contextchanged:.contextsensitiv"
         ajax:action="actionname:#content:replace"
         ajax:confirm="Do you want to perform?"
         ajax:overlay="someaction"
         ajax:path="/foo"
         ajax:path-target="target"
         ajax:path-action="actionname:#content:replace"
         ajax:path-event="contextchanged:.contextsensitiv"
         ajax:path-overlay="someaction"
        >&nbsp;Foo</a>
        """, rendered)

        action.enabled = False
        self.assertTrue(
            action(model, request).find('class="link_action disabled"') > -1
        )

        action.display = False
        self.assertEqual(action(model, request), u'')

    def test_ButtonAction(self):
        model = BaseNode()
        request = self.layer.new_request()

        rendered = ButtonAction()(model, request)
        self.checkOutput("""
        <button
          ajax:bind="click"
          ajax:target="http://example.com/">
        </button>
        """, rendered)

        action = ButtonAction()
        action.id = 'button_id'
        action.css = 'button_class'
        action.title = 'Button Title'
        action.type = 'submit'
        action.name = 'submit-form'
        action.value = 'Submit'
        action.autofocus = 'autofocus'
        action.disabled = 'disabled'
        action.form = '#form-id'
        action.formaction = 'http://example.com/form'
        action.formenctype = 'text/plain'
        action.formmethod = 'post'
        action.formnovalidate = 'formnovalidate'
        action.formtarget = '_blank'
        action.action = 'action:NONE:NONE'
        action.event = 'event:.selector'
        action.confirm = 'Really?'
        action.overlay = 'overlay'
        action.path = '/foo'
        action.path_target = 'target'
        action.path_action = action.action
        action.path_event = action.event
        action.path_overlay = action.overlay
        action.text = 'Button Text'
        action.icon = 'button-icon'

        rendered = action(model, request)
        self.checkOutput("""
        <button id="button_id"
          title="Button Title"
          class="button_class"
          name="submit-form"
          type="submit"
          value="Submit"
          autofocus="autofocus"
          disabled="disabled"
          form="#form-id"
          formaction="http://example.com/form"
          formenctype="text/plain"
          formmethod="post"
          formnovalidate="formnovalidate"
          formtarget="_blank"
          ajax:bind="click"
          ajax:target="http://example.com/"
          ajax:event="event:.selector"
          ajax:action="action:NONE:NONE"
          ajax:confirm="Really?"
          ajax:overlay="overlay"
          ajax:path="/foo"
          ajax:path-target="target"
          ajax:path-action="action:NONE:NONE"
          ajax:path-event="event:.selector"
          ajax:path-overlay="overlay">
          <i class="button-icon"></i>
          <span>Button Text</span>
        </button>
        """, rendered)

        action.display = False
        self.assertEqual(action(model, request), u'')

    def test_ActionUp(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        action = ActionUp()
        self.assertEqual(action(model, request), u'')

        model.properties.action_up = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('viewer'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-up"
            href="http://example.com/root"
            ajax:bind="click"
            ajax:target="http://example.com/root?contenttile=listing"
            ajax:event="contextchanged:#layout"
            ajax:path="href"
            ><span class="glyphicon glyphicon-arrow-up"></span
            >&nbsp;One level up</a>...
            """, rendered)

            model.properties.action_up_tile = 'otherparentcontent'
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-up"
            href="http://example.com/root"
            ajax:bind="click"
            ajax:target="http://example.com/root?contenttile=otherparentcontent"
            ajax:event="contextchanged:#layout"
            ajax:path="href"
            ><span class="glyphicon glyphicon-arrow-up"></span
            >&nbsp;One level up</a>...
            """, rendered)

            default = model['default'] = BaseNode()
            default.properties.action_up = True
            model.properties.default_child = 'default'
            rendered = action(default, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-up"
            href="http://example.com/root"
            ajax:bind="click"
            ajax:target="http://example.com/root?contenttile=listing"
            ajax:event="contextchanged:#layout"
            ajax:path="href"
            ><span class="glyphicon glyphicon-arrow-up"></span
            >&nbsp;One level up</a>...
            """, rendered)

    def test_ActionView(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        ActionContext(model, request, 'content')

        action = ActionView()
        self.assertEqual(action(model, request), u'')

        model.properties.action_view = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('viewer'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-view"
            href="http://example.com/root/model"
            class="selected"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="content:#content:inner"
            ajax:path="href"
            ><span class="glyphicon glyphicon-eye-open"></span
            >&nbsp;View</a>...
            """, rendered)

            model.properties.default_content_tile = 'otherdefault'
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-view"
            href="http://example.com/root/model"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="view:#content:inner"
            ajax:path="href"
            ><span class="glyphicon glyphicon-eye-open"></span
            >&nbsp;View</a>...
            """, rendered)

            model.properties.default_content_tile = None

    def test_ViewLink(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        ActionContext(model, request, 'content')

        action = ViewLink()
        self.assertEqual(action(model, request), u'')

        model.properties.action_view = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('viewer'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-view"
            href="http://example.com/root/model"
            class="selected"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="content:#content:inner"
            ajax:path="href"
            >&nbsp;model</a>...
            """, rendered)

    def test_ActionList(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        action = ActionList()
        self.assertEqual(action(model, request), u'')

        model.properties.action_list = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('viewer'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-list"
            href="http://example.com/root/model/listing"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="listing:#content:inner"
            ajax:path="href"
            ><span class="glyphicon glyphicon-th-list"></span
            >&nbsp;Listing</a>...
            """, rendered)

    def test_ActionSharing(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        action = ActionSharing()
        self.assertFalse(IPrincipalACL.providedBy(model))
        self.assertEqual(action(model, request), u'')

        sharingmodel = parent['sharingmodel'] = SharingNode()
        self.assertTrue(IPrincipalACL.providedBy(sharingmodel))
        self.assertEqual(action(sharingmodel, request), u'')

        with self.layer.authenticated('editor'):
            rule = request.has_permission('manage_permissions', sharingmodel)
            self.assertTrue(isinstance(rule, ACLDenied))
            self.assertEqual(action(sharingmodel, request), u'')

        with self.layer.authenticated('manager'):
            rule = request.has_permission('manage_permissions', sharingmodel)
            self.assertTrue(isinstance(rule, ACLAllowed))

            rendered = action(sharingmodel, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-share"
            href="http://example.com/root/sharingmodel/sharing"
            ajax:bind="click"
            ajax:target="http://example.com/root/sharingmodel"
            ajax:action="sharing:#content:inner"
            ajax:path="href"
            ><span class="glyphicon glyphicon-share"></span
            >&nbsp;Sharing</a>...
            """, rendered)

    def test_ActionState(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        action = ActionState()
        self.assertFalse(IWorkflowState.providedBy(model))
        self.assertEqual(action(model, request), u'')

        wfmodel = parent['wfmodel'] = WorkflowNode()
        self.assertTrue(IWorkflowState.providedBy(wfmodel))
        self.assertEqual(action(wfmodel, request), u'')

        with self.layer.authenticated('editor'):
            rule = request.has_permission('change_state', wfmodel)
            self.assertTrue(isinstance(rule, ACLDenied))
            self.assertEqual(action(wfmodel, request), u'')

        with self.layer.authenticated('manager'):
            rule = request.has_permission('change_state', wfmodel)
            self.assertTrue(isinstance(rule, ACLAllowed))

            rendered = action(wfmodel, request)
            self.checkOutput("""
            ...<li class="dropdown">...
            <a href="#"
            ajax:bind="click"
            ajax:target="http://example.com/root/wfmodel?do_transition=initial_2_final"
            ajax:action="wf_dropdown:NONE:NONE">initial_2_final</a>...
            """, rendered)

    @testing.reset_node_info_registry
    def test_ActionAdd(self):
        info = NodeInfo()
        info.title = 'Addable'
        info.addables = ['addable']
        register_node_info('addable', info)

        action = ActionAdd()
        addmodel = BaseNode()
        request = self.layer.new_request()
        ActionContext(addmodel, request, 'listing')

        self.assertEqual(action(addmodel, request), u'')

        with self.layer.authenticated('viewer'):
            rule = request.has_permission('add', addmodel)
            self.assertTrue(isinstance(rule, ACLDenied))
            self.assertEqual(action(addmodel, request), u'')

        with self.layer.authenticated('editor'):
            rule = request.has_permission('add', addmodel)
            self.assertTrue(isinstance(rule, ACLAllowed))
            self.assertEqual(action(addmodel, request), u'')

            addmodel.node_info_name = 'addable'
            self.assertTrue(addmodel.nodeinfo is info)

            rendered = action(addmodel, request)
            self.checkOutput("""
            ...<li class="dropdown">
            <a href="#"
            class="dropdown-toggle"
            data-toggle="dropdown">
            <span>Add</span>
            <span class="caret"></span>
            </a>
            <ul class="dropdown-menu" role="addmenu">
            <li>
            <a href="http://example.com/add?factory=addable"
            ajax:bind="click"
            ajax:target="http://example.com/?factory=addable"
            ajax:action="add:#content:inner"
            ajax:path="href">
            <span class="glyphicon glyphicon-asterisk"></span>
            Addable
            </a>
            </li>
            </ul>
            </li>...
            """, rendered)

    def test_ActionEdit(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        ActionContext(model, request, 'listing')

        action = ActionEdit()
        self.assertEqual(action(model, request), u'')

        model.properties.action_edit = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('viewer'):
            self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('editor'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-edit"
            href="http://example.com/root/model/edit"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="edit:#content:inner"
            ajax:path="href"
            ><span class="glyphicon glyphicon-pencil"></span
            >&nbsp;Edit</a>...
            """, rendered)

    def test_ActionDelete(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        ActionContext(model, request, 'content')

        action = ActionDelete()
        self.assertEqual(action(model, request), u'')

        model.properties.action_delete = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('editor'):
            self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('manager'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-delete"
            href="#"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="delete:NONE:NONE"
            ajax:confirm="Do you really want to delete this Item?"
            ><span class="ion-trash-a"></span
            >&nbsp;Delete</a>...
            """, rendered)

            model.properties.default_content_tile = 'othertile'
            self.assertEqual(action(model, request), u'')

    def test_ActionDeleteChildren(self):
        parent = BaseNode(name='root')
        model = parent['model'] = BaseNode()
        request = self.layer.new_request()

        action = ActionDeleteChildren()
        self.assertEqual(action(model, request), u'')

        model.properties.action_delete_children = True
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('editor'):
            self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('manager'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-delete-children"
            href="#"
            class="disabled"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="delete_children:NONE:NONE"
            ajax:confirm="Do you really want to delete selected Items?"
            ><span class="ion-trash-a"></span
            >&nbsp;Delete selected children</a>...
            """, rendered)

            request.cookies['cone.app.selected'] = ['foo']
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-delete-children"
            href="#"
            ajax:bind="click"
            ajax:target="http://example.com/root/model"
            ajax:action="delete_children:NONE:NONE"
            ajax:confirm="Do you really want to delete selected Items?"
            ><span class="ion-trash-a"></span
            >&nbsp;Delete selected children</a>...
            """, rendered)

            del request.cookies['cone.app.selected']

    def test_ActionCut(self):
        model = CopySupportNode('copysupport')
        request = self.layer.new_request()

        ActionContext(model, request, 'listing')
        self.assertTrue(ICopySupport.providedBy(model))
        self.assertTrue(model.supports_cut)

        action = ActionCut()
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('editor'):
            self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('manager'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-cut"
            href="#"
            ajax:target="http://example.com/copysupport"
            ><span class="ion-scissors"></span
            >&nbsp;Cut</a>...
            """, rendered)

            model.supports_cut = False
            self.assertEqual(action(model, request), u'')

    def test_ActionCopy(self):
        model = CopySupportNode('copysupport')
        request = self.layer.new_request()

        ActionContext(model, request, 'listing')
        self.assertTrue(model.supports_copy)

        action = ActionCopy()
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('editor'):
            self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('manager'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-copy"
            href="#"
            ajax:target="http://example.com/copysupport"
            ><span class="ion-ios7-copy-outline"></span
            >&nbsp;Copy</a>...
            """, rendered)

            model.supports_copy = False
            self.assertEqual(action(model, request), u'')

    def test_ActionPaste(self):
        model = CopySupportNode('copysupport')
        request = self.layer.new_request()

        ActionContext(model, request, 'listing')
        self.assertTrue(model.supports_paste)

        action = ActionPaste()
        self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('editor'):
            self.assertEqual(action(model, request), u'')

        with self.layer.authenticated('manager'):
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-paste"
            href="#"
            class="disabled"
            ajax:target="http://example.com/copysupport"
            ><span class="ion-clipboard"></span
            >&nbsp;Paste</a>...
            """, rendered)

            request.cookies['cone.app.copysupport.cut'] = ['foo']
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-paste"
            href="#"
            ajax:target="http://example.com/copysupport"
            ><span class="ion-clipboard"></span
            >&nbsp;Paste</a>...
            """, rendered)

            del request.cookies['cone.app.copysupport.cut']
            request.cookies['cone.app.copysupport.copy'] = ['foo']
            rendered = action(model, request)
            self.checkOutput("""
            ...<a
            id="toolbaraction-paste"
            href="#"
            ajax:target="http://example.com/copysupport"
            ><span class="ion-clipboard"></span
            >&nbsp;Paste</a>...
            """, rendered)

            del request.cookies['cone.app.copysupport.copy']

            model.supports_paste = False
            self.assertEqual(action(model, request), u'')

    def test__ActionMove(self):
        node = BaseNode()
        model = node['child'] = BaseNode()
        request = self.layer.new_request()

        action = _ActionMove()
        action.model = model
        action.request = request

        request.params['sort'] = 'asc'
        self.assertFalse(action.display)

        del request.params['sort']
        node.properties.action_move = True
        self.assertFalse(action.display)

        @plumbing(Order)
        class OrderableNode(BaseNode):
            pass

        node = OrderableNode()
        node.properties.action_move = True
        model = node['child'] = BaseNode()
        action.model = model
        with self.layer.authenticated('manager'):
            self.assertTrue(action.display)

        self.assertEqual(action.target, 'http://example.com/child')
        request.params['size'] = '10'
        request.params['b_page'] = '1'
        self.assertEqual(
            action.target,
            'http://example.com/child?b_page=1&size=10'
        )

    def test_ActionMoveUp(self):
        action = ActionMoveUp()
        self.assertEqual(action.id, 'toolbaraction-move-up')
        self.assertEqual(action.icon, 'glyphicon glyphicon-chevron-up')
        self.assertEqual(action.action, 'move_up:NONE:NONE')
        self.assertEqual(action.text, 'move_up')

        node = BaseNode()
        model = node['child'] = BaseNode()
        request = self.layer.new_request()

        action.model = model
        action.request = request
        self.assertFalse(action.display)

        @plumbing(Order)
        class OrderableNode(BaseNode):
            pass

        node = OrderableNode()
        node.properties.action_move = True
        node['a'] = BaseNode()
        node['b'] = BaseNode()

        action.model = node['a']
        with self.layer.authenticated('manager'):
            self.assertFalse(action.display)

        action.model = node['b']
        with self.layer.authenticated('manager'):
            self.assertTrue(action.display)

    def test_ActionMoveDown(self):
        action = ActionMoveDown()
        self.assertEqual(action.id, 'toolbaraction-move-down')
        self.assertEqual(action.icon, 'glyphicon glyphicon-chevron-down')
        self.assertEqual(action.action, 'move_down:NONE:NONE')
        self.assertEqual(action.text, 'move_down')

        node = BaseNode()
        model = node['child'] = BaseNode()
        request = self.layer.new_request()

        action.model = model
        action.request = request
        self.assertFalse(action.display)

        @plumbing(Order)
        class OrderableNode(BaseNode):
            pass

        node = OrderableNode()
        node.properties.action_move = True
        node['a'] = BaseNode()
        node['b'] = BaseNode()

        action.model = node['a']
        with self.layer.authenticated('manager'):
            self.assertTrue(action.display)

        action.model = node['b']
        with self.layer.authenticated('manager'):
            self.assertFalse(action.display)
